# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   (c) 2019 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Modules that define classes used for unit testing the workbench.

These modules contain classes and functions that are called
by the unit test module that is defined in `Init.py` and `InitGui.py`.

The unit tests are placed in separate modules in order to test features
that do not require the graphical user interface (GUI), from those
that do require it.

The unit tests are based on the standard Python `unittest` module.
See this module and `unittest.TestCase` for more information
on how to write unit tests.

::

    class NewTestType(unittest.TestCase):
        def test_new_tool(self):
            pass
"""
## \defgroup drafttests drafttests
# \ingroup DRAFT
# \brief Modules that define classes used for unit testing the workbench.



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2023 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""Contains a parameter observer class and parameter related functions."""

import os
import PySide.QtCore as QtCore
import xml.etree.ElementTree as ET

import FreeCAD as App
import Draft_rc

try:
    import Arch_rc
except ModuleNotFoundError:
    pass

from draftutils.translate import translate

if App.GuiUp:
    import FreeCADGui as Gui
    from PySide import QtWidgets


class ParamObserverDraft:

    def slotParamChanged(self, param_grp, typ, entry, value):
        if entry == "textheight":
            _param_observer_callback_tray()
            return
        if entry in (
            "gridBorder",
            "gridShowHuman",
            "coloredGridAxes",
            "gridEvery",
            "gridSpacing",
            "gridSize",
            "gridTransparency",
            "gridColor",
        ):
            _param_observer_callback_grid()
            return
        if entry == "DefaultAnnoScaleMultiplier":
            _param_observer_callback_scalemultiplier(value)
            return
        if entry == "SnapBarShowOnlyDuringCommands":
            _param_observer_callback_snapbar(value)
            return
        if entry == "DisplayStatusbarSnapWidget":
            _param_observer_callback_snapwidget()
            return
        if entry == "DisplayStatusbarScaleWidget":
            _param_observer_callback_scalewidget()
            return
        if entry == "snapStyle":
            _param_observer_callback_snapstyle()
            return
        if entry == "snapcolor":
            _param_observer_callback_snapcolor()
            return
        if entry == "patternFile":
            _param_observer_callback_svg_pattern()
            return


class ParamObserverView:

    def slotParamChanged(self, param_grp, typ, entry, value):
        if entry in ("DefaultShapeColor", "DefaultShapeLineColor", "DefaultShapeLineWidth"):
            _param_observer_callback_tray()
            return
        if entry == "MarkerSize":
            _param_observer_callback_snaptextsize()
            return


def _param_observer_callback_tray():
    if not hasattr(Gui, "draftToolBar"):
        return
    tray = Gui.draftToolBar.tray
    if tray is None:
        return
    Gui.draftToolBar.setStyleButton()


def _param_observer_callback_scalemultiplier(value):
    # value is a string.
    # import has to happen here to avoid circular imports
    from draftutils import init_draft_statusbar

    if not value:
        return
    value = float(value)
    if value <= 0:
        return
    mw = Gui.getMainWindow()
    sb = mw.statusBar()
    scale_widget = sb.findChild(QtWidgets.QToolBar, "draft_scale_widget")
    if scale_widget is not None:
        scale_label = init_draft_statusbar.scale_to_label(1 / value)
        scale_widget.scaleLabel.setText(scale_label)


def _param_observer_callback_grid():
    if hasattr(App, "draft_working_planes") and hasattr(Gui, "Snapper"):
        try:
            trackers = Gui.Snapper.trackers
            for wp in App.draft_working_planes[1]:
                view = wp._view
                if view in trackers[0]:
                    i = trackers[0].index(view)
                    grid = trackers[1][i]
                    grid.pts = []
                    grid.reset()
                    grid.displayHumanFigure(wp)
                    grid.setAxesColor(wp)
        except Exception:
            pass


def _param_observer_callback_snapbar(value):
    if Gui.activeWorkbench().name() not in ("DraftWorkbench", "ArchWorkbench", "BIMWorkbench"):
        return
    if hasattr(Gui, "Snapper"):
        toolbar = Gui.Snapper.get_snap_toolbar()
        if toolbar is not None:
            toolbar.setVisible(value == "0")  # value is a string: "0" or "1"


def _param_observer_callback_snapwidget():
    # import has to happen here to avoid circular imports
    from draftutils import init_draft_statusbar

    if Gui.activeWorkbench().name() == "DraftWorkbench":
        init_draft_statusbar.hide_draft_statusbar()
        init_draft_statusbar.show_draft_statusbar()


def _param_observer_callback_scalewidget():
    # import has to happen here to avoid circular imports
    from draftutils import init_draft_statusbar

    if Gui.activeWorkbench().name() == "DraftWorkbench":
        init_draft_statusbar.hide_draft_statusbar()
        init_draft_statusbar.show_draft_statusbar()


def _param_observer_callback_snapstyle():
    if hasattr(Gui, "Snapper"):
        Gui.Snapper.set_snap_style()


def _param_observer_callback_snapcolor():
    if hasattr(Gui, "Snapper"):
        tracker_list = [2, 5, 6]
        for each_tracker in tracker_list:
            for snap_track in Gui.Snapper.trackers[each_tracker]:
                snap_track.setColor()


def _param_observer_callback_snaptextsize():
    if hasattr(Gui, "Snapper"):
        tracker_list = [5, 6]
        for each_tracker in tracker_list:
            for snap_track in Gui.Snapper.trackers[each_tracker]:
                snap_track.setSize()


def _param_observer_callback_svg_pattern():
    # imports have to happen here to avoid circular imports
    from draftutils import utils
    from draftviewproviders import view_base

    utils.load_svg_patterns()
    if App.ActiveDocument is None:
        return

    pats = list(utils.svg_patterns())
    pats.sort()
    pats = ["None"] + pats

    data = []
    for doc in App.listDocuments().values():
        vobjs = []
        for obj in doc.Objects:
            if hasattr(obj, "ViewObject"):
                vobj = obj.ViewObject
                if (
                    hasattr(vobj, "Pattern")
                    and hasattr(vobj, "Proxy")
                    and isinstance(vobj.Proxy, view_base.ViewProviderDraft)
                    and vobj.getEnumerationsOfProperty("Pattern") != pats
                ):
                    vobjs.append(vobj)
        if vobjs:
            data.append([doc, vobjs])
    if not data:
        return

    msg = translate(
        "draft",
        """Do you want to update the SVG pattern options
of existing objects in all opened documents?""",
    )
    res = QtWidgets.QMessageBox.question(
        None,
        "Update SVG patterns",
        msg,
        QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
        QtWidgets.QMessageBox.No,
    )
    if res == QtWidgets.QMessageBox.No:
        return

    for doc, vobjs in data:
        doc.openTransaction("SVG pattern update")
        for vobj in vobjs:
            old = vobj.Pattern
            if old in pats:
                vobj.Pattern = pats
            else:
                tmp_pats = [old] + pats[1:]
                tmp_pats.sort()
                vobj.Pattern = ["None"] + tmp_pats
            vobj.Pattern = old
        doc.commitTransaction()


def _param_observer_start():
    if App.GuiUp:
        _param_observer_start_draft()
        _param_observer_start_view()


def _param_observer_start_draft(
    param_grp=App.ParamGet("User parameter:BaseApp/Preferences/Mod/Draft"),
):
    param_grp.AttachManager(ParamObserverDraft())


def _param_observer_start_view(param_grp=App.ParamGet("User parameter:BaseApp/Preferences/View")):
    param_grp.AttachManager(ParamObserverView())


def _param_from_PrefCheckBox(widget):
    value = False
    for elem in list(widget):
        if "name" in elem.keys():
            att_name = elem.attrib["name"]
            if att_name == "checked":  # Can be missing.
                value = elem.find("bool").text == "true"
            elif att_name == "prefEntry":
                entry = elem.find("cstring").text
            elif att_name == "prefPath":
                path = elem.find("cstring").text
    return path, entry, value


def _param_from_PrefRadioButton(widget):
    return _param_from_PrefCheckBox(widget)


def _param_from_PrefComboBox(widget):
    value = 0
    for elem in list(widget):
        if "name" in elem.keys():
            att_name = elem.attrib["name"]
            if att_name == "currentIndex":  # Can be missing.
                value = int(elem.find("number").text)
            elif att_name == "prefEntry":
                entry = elem.find("cstring").text
            elif att_name == "prefPath":
                path = elem.find("cstring").text
    return path, entry, value


def _param_from_PrefSpinBox(widget):
    value = 0
    for elem in list(widget):
        if "name" in elem.keys():
            att_name = elem.attrib["name"]
            if att_name == "value":  # Can be missing.
                value = int(elem.find("number").text)
            elif att_name == "prefEntry":
                entry = elem.find("cstring").text
            elif att_name == "prefPath":
                path = elem.find("cstring").text
    return path, entry, value


def _param_from_PrefDoubleSpinBox(widget):
    value = 0.0
    for elem in list(widget):
        if "name" in elem.keys():
            att_name = elem.attrib["name"]
            if att_name == "value":  # Can be missing.
                value = float(elem.find("double").text)
            elif att_name == "prefEntry":
                entry = elem.find("cstring").text
            elif att_name == "prefPath":
                path = elem.find("cstring").text
    return path, entry, value


def _param_from_PrefUnitSpinBox(widget):
    value = 0.0
    for elem in list(widget):
        if "name" in elem.keys():
            att_name = elem.attrib["name"]
            if att_name == "rawValue":  # Can be missing.
                value = float(elem.find("double").text)
            elif att_name == "prefEntry":
                entry = elem.find("cstring").text
            elif att_name == "prefPath":
                path = elem.find("cstring").text
    return path, entry, value


def _param_from_PrefQuantitySpinBox(widget):
    value = "0"
    for elem in list(widget):
        if "name" in elem.keys():
            att_name = elem.attrib["name"]
            if att_name == "rawValue":  # Can be missing.
                value = elem.find("double").text.rstrip(".0")
            elif att_name == "unit":
                unit = elem.find("string").text
            elif att_name == "prefEntry":
                entry = elem.find("cstring").text
            elif att_name == "prefPath":
                path = elem.find("cstring").text
    value = value + " " + unit
    return path, entry, value


def _param_from_PrefColorButton(widget):
    for elem in list(widget):
        if "name" in elem.keys():
            att_name = elem.attrib["name"]
            if att_name == "color":
                sub = list(elem)[0]
                r = int(sub.find("red").text)
                g = int(sub.find("green").text)
                b = int(sub.find("blue").text)
            elif att_name == "prefEntry":
                entry = elem.find("cstring").text
            elif att_name == "prefPath":
                path = elem.find("cstring").text
    value = (r << 24) + (g << 16) + (b << 8) + 255
    return path, entry, value


def _param_from_PrefLineEdit(widget):
    value = None
    for elem in list(widget):
        if "name" in elem.keys():
            att_name = elem.attrib["name"]
            if att_name == "text":  # Can be missing.
                value = elem.find("string").text  # If text is missing value will be None here.
            elif att_name == "prefEntry":
                entry = elem.find("cstring").text
            elif att_name == "prefPath":
                path = elem.find("cstring").text
    if value is None:
        value = ""
    return path, entry, value


def _param_from_PrefFileChooser(widget):
    # Does not occur in Draft preferences anymore.
    for elem in list(widget):
        if "name" in elem.keys():
            att_name = elem.attrib["name"]
            if att_name == "prefEntry":
                entry = elem.find("cstring").text
            elif att_name == "prefPath":
                path = elem.find("cstring").text
    return path, entry, ""


def _param_from_PrefFontBox(widget):
    if App.GuiUp:
        from PySide import QtGui

        font = QtGui.QFont()
        font.setStyleHint(QtGui.QFont.StyleHint.SansSerif)
        value = font.defaultFamily()
    else:
        value = ""
    for elem in list(widget):
        if "name" in elem.keys():
            att_name = elem.attrib["name"]
            if att_name == "prefEntry":
                entry = elem.find("cstring").text
            elif att_name == "prefPath":
                path = elem.find("cstring").text
    # We must set the parameter if it does not exist, else
    # the Gui::PrefFontBox will show the wrong value.
    param_grp = App.ParamGet("User parameter:BaseApp/Preferences/" + path)
    if entry not in param_grp.GetStrings():
        param_grp.SetString(entry, value)
    return path, entry, value


def _get_shape_string_font_file():
    """Try to get the file name of a sans serif font (TTC or TTF) from the OS."""
    # https://forum.freecad.org/viewtopic.php?t=96770
    # Windows font: "arial"
    # Mac fonts: "geneva" and "helvetica"
    # Linux fonts: "dejavusans" and "freesans"
    favorite_names = ("arial", "geneva", "helvetica", "dejavusans", "freesans")
    font_file_sans = None  # Font with name containing "sans". 1st fallback.
    font_file_alpha = None  # Font with name starting with a letter. 2nd fallback.
    # Reverse the order of the paths so that user related paths come last:
    for path in QtCore.QStandardPaths.standardLocations(QtCore.QStandardPaths.FontsLocation)[::-1]:
        # We don't use os.path.join as dir_path has forward slashes even on Windows.
        for dir_path, dir_names, file_names in os.walk(path):
            for file_name in file_names:
                base_name, ext = [s.lower() for s in os.path.splitext(file_name)]
                if not ext in (".ttc", ".ttf"):
                    continue
                if base_name in favorite_names:
                    return dir_path + "/" + file_name
                if font_file_sans is None and "sans" in base_name:
                    font_file_sans = dir_path + "/" + file_name
                if font_file_alpha is None and base_name[0].isalpha():
                    font_file_alpha = dir_path + "/" + file_name
    if font_file_sans is not None:
        return font_file_sans
    if font_file_alpha is not None:
        return font_file_alpha
    return ""


def _get_param_dictionary():

    # print("Creating preferences dictionary...")

    param_dict = {}

    hatch_pattern_file = (
        App.getResourceDir().replace("\\", "/").rstrip("/") + "/Mod/TechDraw/PAT/FCPAT.pat"
    )

    # Draft parameters that are not in the preferences:
    # fmt: off
    param_dict["Mod/Draft"] = {
        "AnnotationStyleEditorHeight": ("int",       450),
        "AnnotationStyleEditorWidth":  ("int",       600),
        "CenterPlaneOnView":           ("bool",      False),
        "ChainedMode":                 ("bool",      False),
        "CopyMode":                    ("bool",      False),
        "DefaultAnnoDisplayMode":      ("int",       0),
        "DefaultDisplayMode":          ("int",       0),
        "DefaultDrawStyle":            ("int",       0),
        "DefaultPrintColor":           ("unsigned",  255),
        "DimAutoFlipText":             ("bool",      True),
        "Draft_array_fuse":            ("bool",      False),
        "Draft_array_Link":            ("bool",      True),
        "FilletChamferMode":           ("bool",      False),
        "FilletDeleteMode":            ("bool",      False),
        "FilletRadius":                ("float",     100.0),
        "GlobalMode":                  ("bool",      False),
        "GridHideInOtherWorkbenches":  ("bool",      True),
        "HatchPatternFile":            ("string",    hatch_pattern_file),
        "HatchPatternName":            ("string",    "Diamond"),
        "HatchPatternResolution":      ("int",       128),  # used for SVG patterns
        "HatchPatternRotation":        ("float",     0.0),
        "HatchPatternScale":           ("float",     100.0),
        "HatchPatternTranslate":       ("bool",      True),
        "labeltype":                   ("string",    "Custom"),
        "LayersManagerHeight":         ("int",       320),
        "LayersManagerWidth":          ("int",       640),
        "MakeFaceMode":                ("bool",      True),
        "maxSnapEdges":                ("int",       0),
        "OffsetCopyMode":              ("bool",      False),
        "Offset_OCC":                  ("bool",      False),
        "RelativeMode":                ("bool",      True),
        "ScaleClone":                  ("bool",      False),
        "ScaleCopy":                   ("bool",      False),
        "ScaleRelative":               ("bool",      False),
        "ScaleUniform":                ("bool",      False),
        "ShapeStringFontFile":         ("string",    _get_shape_string_font_file()),
        "ShapeStringHeight":           ("float",     10.0),
        "ShapeStringText":             ("string",    translate("draft", "Default")),
        "snapModes":                   ("string",    "100000000000000"),
        "snapRange":                   ("int",       8),
        "SubelementMode":              ("bool",      False),
        "SvgLinesBlack":               ("bool",      True),
        "useSupport":                  ("bool",      False),
    }

    param_dict["Mod/Draft/ContinueMode"] = {
        # Draft
        "Arc":                         ("bool",      False),
        "Arc_3Points":                 ("bool",      False),
        "BezCurve":                    ("bool",      False),
        "Bspline":                     ("bool",      False),
        "Circle":                      ("bool",      False),
        "CubicBezCurve":               ("bool",      False),
        "Dimension":                   ("bool",      False),
        "Ellipse":                     ("bool",      False),
        "Line":                        ("bool",      False),
        "Point":                       ("bool",      False),
        "Polygon":                     ("bool",      False),
        "Polyline":                    ("bool",      False),
        "Rectangle":                   ("bool",      False),
        "Text":                        ("bool",      False),

        # Standard operations (Draft)
        "Copy":                        ("bool",      False),
        "Move":                        ("bool",      False),
        "Rotate":                      ("bool",      False),

        # Arch/BIM
        "Beam":                        ("bool",      False),
        "Column":                      ("bool",      False),
        "Panel":                       ("bool",      False),
        "Wall":                        ("bool",      False),
    }

    start_val = App.Units.Quantity(100.0, App.Units.Length).Value
    param_dict["Mod/Draft/OrthoArrayLinearMode"] = {
        "LinearModeOn":                ("bool",      True),
        "AxisSelected":                ("string",    "X"),
        "XInterval":                   ("float",     start_val),
        "YInterval":                   ("float",     start_val),
        "ZInterval":                   ("float",     start_val),
        "XNumOfElements":              ("int",       2),
        "YNumOfElements":              ("int",       2),
        "ZNumOfElements":              ("int",       2)
    }

    # Arch parameters that are not in the preferences:
    param_dict["Mod/Arch"] = {
        "applyConstructionStyle":      ("bool",      True),
        "ClaimHosted":                 ("bool",      True),
        "CustomIfcSchema":             ("string",    ""),     # importIFClegacy.py
        "createIfcGroups":             ("bool",      False),  # importIFClegacy.py
        "DoorHeight":                  ("float",     2100.0),
        "DoorPreset":                  ("int",       5),
        "DoorSill":                    ("float",     0.0),
        "DoorWidth":                   ("float",     1000.0),
        "FreeLinking":                 ("bool",      False),
        "forceIfcPythonParser":        ("bool",      False),  # importIFClegacy.py
        "getStandardType":             ("bool",      False),
        "ifcAggregateWindows":         ("bool",      False),  # importIFClegacy.py
        "ifcAsMesh":                   ("string",    ""),     # importIFClegacy.py
        "IfcExportList":               ("bool",      False),  # importIFClegacy.py
        "ifcImportLayer":              ("bool",      True),
        "ifcJoinSolids":               ("bool",      False),  # importIFClegacy.py
        "ifcMergeProfiles":            ("bool",      False),
        "IfcScalingFactor":            ("float",     1.0),    # importIFClegacy.py
        "ifcSeparatePlacements":       ("bool",      False),  # importIFClegacy.py
        "MultiMaterialColumnWidth0":   ("int",       120),
        "MultiMaterialColumnWidth1":   ("int",       120),
        "PanelLength":                 ("float",     1000.0),
        "PanelThickness":              ("float",     10.0),
        "PanelWidth":                  ("float",     1000.0),
        "PrecastBase":                 ("float",     0.0),
        "PrecastChamfer":              ("float",     0.0),
        "PrecastDentHeight":           ("float",     0.0),
        "PrecastDentLength":           ("float",     0.0),
        "PrecastDentWidth":            ("float",     0.0),
        "PrecastDownLength":           ("float",     0.0),
        "PrecastGrooveDepth":          ("float",     0.0),
        "PrecastGrooveHeight":         ("float",     0.0),
        "PrecastGrooveSpacing":        ("float",     0.0),
        "PrecastHoleMajor":            ("float",     0.0),
        "PrecastHoleMinor":            ("float",     0.0),
        "PrecastHoleSpacing":          ("float",     0.0),
        "PrecastRiser":                ("float",     0.0),
        "PrecastTread":                ("float",     0.0),
        "ProfilePreset":               ("string",    ""),
        "ScheduleColumnWidth0":        ("int",       100),
        "ScheduleColumnWidth1":        ("int",       100),
        "ScheduleColumnWidth2":        ("int",       50),
        "ScheduleColumnWidth3":        ("int",       100),
        "ScheduleDialogHeight":        ("int",       200),
        "ScheduleDialogWidth":         ("int",       300),
        "StructureHeight":             ("float",     1000.0),
        "StructureLength":             ("float",     100.0),
        "StructurePreset":             ("string",    ""),
        "StructureWidth":              ("float",     100.0),
        "swallowAdditions":            ("bool",      True),
        "swallowSubtractions":         ("bool",      True),
        "WallAlignment":               ("int",       0),
        "WallHeight":                  ("float",     3000.0),
        "WallWidth":                   ("float",     200.0),
        "WallOffset":                  ("float",     0.0),
        "WindowH1":                    ("float",     50.0),
        "WindowH2":                    ("float",     50.0),
        "WindowH3":                    ("float",     50.0),
        "WindowHeight":                ("float",     1000.0),
        "WindowO1":                    ("float",     0.0),
        "WindowO2":                    ("float",     50.0),
        "WindowPreset":                ("int",       0),
        "WindowSill":                  ("float",     0.0),
        "WindowW1":                    ("float",     100.0),
        "WindowW2":                    ("float",     50.0),
        "WindowWidth":                 ("float",     1000.0),
    }

    # BIM parameters that are not in the preferences:
    # Note: incomplete!
    param_dict["Mod/BIM"] = {
        "BIMSketchPlacementOnly":      ("bool",      False),
    }

    # For the Mod/Mesh parameters we do not check the preferences:
    param_dict["Mod/Mesh"] = {
        "MaxDeviationExport":          ("float",     0.1),
    }

    # For the General parameters we do not check the preferences:
    param_dict["General"] = {
        "ToolbarIconSize":             ("int",       24),
    }

    # For the Units parameters we do not check the preferences:
    param_dict["Units"] = {
        "Decimals":                    ("int",       2),
        "UserSchema":                  ("int",       0),
    }

    # For the View parameters we do not check the preferences:
    param_dict["View"] = {
        "BackgroundColor":             ("unsigned",  336897023),
        "BackgroundColor2":            ("unsigned",  859006463),
        "BackgroundColor3":            ("unsigned",  2543299327),
        "DefaultAmbientColor":         ("unsigned",  1431655935),
        "DefaultEmissiveColor":        ("unsigned",  255),
        "DefaultShapeColor":           ("unsigned",  3435980543),
        "DefaultShapeLineColor":       ("unsigned",  421075455),
        "DefaultShapeLineWidth":       ("int",       2),
        "DefaultShapePointSize":       ("int",       2),
        "DefaultShapeShininess":       ("int",       90),
        "DefaultShapeTransparency":    ("int",       0),
        "DefaultShapeVertexColor":     ("unsigned",  421075455),
        "DefaultSpecularColor":        ("unsigned",  2290649343),
        "EnableSelection":             ("bool",      True),
        "Gradient":                    ("bool",      True),
        "MarkerSize":                  ("int",       9),
        "NewDocumentCameraScale":      ("float",     100.0),
    }
    # fmt: on

    # Preferences ui files are stored in resource files.
    # For the Draft Workbench: /Mod/Draft/Draft_rc.py
    # For the Arch Workbench: /Mod/Arch/Arch_rc.py
    for fnm in (
        ":/ui/preferences-draft.ui",
        ":/ui/preferences-draftinterface.ui",
        ":/ui/preferences-draftsnap.ui",
        ":/ui/preferences-drafttexts.ui",
        ":/ui/preferences-draftvisual.ui",
        ":/ui/preferences-dwg.ui",
        ":/ui/preferences-dxf.ui",
        ":/ui/preferences-oca.ui",
        ":/ui/preferences-svg.ui",
        ":/ui/preferences-arch.ui",
        ":/ui/preferences-archdefaults.ui",
        ":/ui/preferences-dae.ui",
        ":/ui/preferences-ifc.ui",
        ":/ui/preferences-ifc-export.ui",
        ":/ui/preferences-sh3d-import.ui",
        ":/ui/preferences-webgl.ui",
    ):

        # https://stackoverflow.com/questions/14750997/load-txt-file-from-resources-in-python
        fd = QtCore.QFile(fnm)
        if fd.open(QtCore.QIODevice.ReadOnly | QtCore.QFile.Text):
            text = QtCore.QTextStream(fd).readAll()
            fd.close()
        else:
            continue

        # https://docs.python.org/3/library/xml.etree.elementtree.html
        root = ET.fromstring(text)

        # Get all preference widgets:
        # pref_widgets = [wid for wid in root.iter("widget") if "Gui::Pref" in wid.attrib["class"]]

        for widget in root.iter("widget"):
            if "class" in widget.keys():
                path = None
                att_class = widget.attrib["class"]

                if att_class == "Gui::PrefCheckBox":
                    path, entry, value = _param_from_PrefCheckBox(widget)
                    typ = "bool"
                elif att_class == "Gui::PrefRadioButton":
                    path, entry, value = _param_from_PrefRadioButton(widget)
                    typ = "bool"
                elif att_class == "Gui::PrefComboBox":
                    path, entry, value = _param_from_PrefComboBox(widget)
                    typ = "int"
                elif att_class == "Gui::PrefSpinBox":
                    path, entry, value = _param_from_PrefSpinBox(widget)
                    typ = "int"
                elif att_class == "Gui::PrefDoubleSpinBox":
                    path, entry, value = _param_from_PrefDoubleSpinBox(widget)
                    typ = "float"
                elif att_class == "Gui::PrefUnitSpinBox":
                    path, entry, value = _param_from_PrefUnitSpinBox(widget)
                    typ = "float"
                elif att_class == "Gui::PrefQuantitySpinBox":
                    path, entry, value = _param_from_PrefQuantitySpinBox(widget)
                    typ = "string"
                elif att_class == "Gui::PrefColorButton":
                    path, entry, value = _param_from_PrefColorButton(widget)
                    typ = "unsigned"
                elif att_class == "Gui::PrefLineEdit":
                    path, entry, value = _param_from_PrefLineEdit(widget)
                    typ = "string"
                elif att_class == "Gui::PrefFileChooser":
                    path, entry, value = _param_from_PrefFileChooser(widget)
                    typ = "string"
                elif att_class == "Gui::PrefFontBox":
                    path, entry, value = _param_from_PrefFontBox(widget)
                    typ = "string"
                elif att_class == "Gui::PrefCheckableGroupBox":
                    # It's a boolean preference, so we can reuse the parsing logic
                    # from _param_from_PrefCheckBox, which looks for <property name="checked">.
                    path, entry, value = _param_from_PrefCheckBox(widget)
                    typ = "bool"

                if path is not None:
                    if path in param_dict:
                        param_dict[path][entry] = (typ, value)
                    else:
                        param_dict[path] = {entry: (typ, value)}

    return param_dict


PARAM_DICT = _get_param_dictionary()


def get_param(entry, path="Mod/Draft", ret_default=False, silent=False):
    """Return a stored parameter value or its default.

    Parameters
    ----------
    entry: str
        Name of the parameter.
    path: str, optional
        Defaults to "Mod/Draft".
        The path where the parameter can be found.
        This string is appended to "User parameter:BaseApp/Preferences/".
    ret_default: bool, optional
        Defaults to `False`.
        If `True`, always return the default value even if a stored value is available.
    silent: bool, optional
        Defaults to `False`.
        If `True`, do not log anything if entry wasn't found.

    Returns
    -------
    bool, float, int or str (if successful) or `None`.
    """
    if path not in PARAM_DICT or entry not in PARAM_DICT[path]:
        if not silent:
            print(f"draftutils.params.get_param: Unable to find '{entry}' in '{path}'")
        return None
    param_grp = App.ParamGet("User parameter:BaseApp/Preferences/" + path)
    typ, default = PARAM_DICT[path][entry]
    if ret_default:
        return default
    if typ == "bool":
        return param_grp.GetBool(entry, default)
    if typ == "float":
        return param_grp.GetFloat(entry, default)
    if typ == "int":
        return param_grp.GetInt(entry, default)
    if typ == "string":
        return param_grp.GetString(entry, default)
    if typ == "unsigned":
        return param_grp.GetUnsigned(entry, default)
    return None


def get_param_arch(entry, ret_default=False):
    return get_param(entry, path="Mod/Arch", ret_default=ret_default)


def get_param_view(entry, ret_default=False):
    return get_param(entry, path="View", ret_default=ret_default)


def set_param(entry, value, path="Mod/Draft"):
    """Store a parameter value.

    Parameters
    ----------
    entry: str
        Name of the parameter.
    value: bool, float, int or str
        New value of the correct type.
    path: str, optional
        Defaults to "Mod/Draft".
        The path where the parameter can be found.
        This string is appended to "User parameter:BaseApp/Preferences/".

    Returns
    -------
    `True` (if successful) or `False`.
    """
    if path not in PARAM_DICT or entry not in PARAM_DICT[path]:
        print(f"draftutils.params.set_param: Unable to find '{entry}' in '{path}'")
        return False
    param_grp = App.ParamGet("User parameter:BaseApp/Preferences/" + path)
    typ = PARAM_DICT[path][entry][0]
    ret = True
    if typ == "bool":
        param_grp.SetBool(entry, value)
    elif typ == "float":
        param_grp.SetFloat(entry, value)
    elif typ == "int":
        param_grp.SetInt(entry, value)
    elif typ == "string":
        param_grp.SetString(entry, value)
    elif typ == "unsigned":
        param_grp.SetUnsigned(entry, value)
    else:
        ret = False
    return ret


def set_param_arch(entry, value):
    return set_param(entry, value, path="Mod/Arch")


def set_param_view(entry, value):
    return set_param(entry, value, path="View")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2023 Robert Schöftner <rs@unfoo.net>                    *
# *   Copyright (c) 2021 Russell Johnson (russ4262) <russ4262@gmail.com>    *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Part
import Path.Op.Profile as PathProfile
import Path.Main.Job as PathJob
from CAMTests.PathTestUtils import PathTestBase
from CAMTests.TestPathAdaptive import getGcodeMoves

if FreeCAD.GuiUp:
    import Path.Main.Gui.Job as PathJobGui
    import Path.Op.Gui.Profile as PathProfileGui


class TestPathProfile(PathTestBase):
    """Unit tests for the Adaptive operation."""

    @classmethod
    def setUpClass(cls):
        """setUpClass()...
        This method is called upon instantiation of this test class.  Add code and objects here
        that are needed for the duration of the test() methods in this class.  In other words,
        set up the 'global' test environment here; use the `setUp()` method to set up a 'local'
        test environment.
        This method does not have access to the class `self` reference, but it
        is able to call static methods within this same class.
        """
        cls.needsInit = True

    @classmethod
    def initClass(cls):
        # Open existing FreeCAD document with test geometry
        cls.needsInit = False
        FreeCAD.ConfigSet("SuppressRecomputeRequiredDialog", "True")
        cls.doc = FreeCAD.open(FreeCAD.getHomePath() + "Mod/CAM/CAMTests/test_profile.fcstd")
        FreeCAD.ConfigSet("SuppressRecomputeRequiredDialog", "")

        # Create Job object, adding geometry objects from file opened above
        cls.job = PathJob.Create("Job", [cls.doc.Body], None)
        cls.job.GeometryTolerance.Value = 0.001
        if FreeCAD.GuiUp:
            cls.job.ViewObject.Proxy = PathJobGui.ViewProvider(cls.job.ViewObject)

        # Instantiate an Profile operation for querying available properties
        cls.prototype = PathProfile.Create("Profile")
        cls.prototype.Base = [(cls.doc.Body, ["Face18"])]
        cls.prototype.Label = "Prototype"
        _addViewProvider(cls.prototype)

        cls.doc.recompute()

    @classmethod
    def tearDownClass(cls):
        """tearDownClass()...
        This method is called prior to destruction of this test class.  Add code and objects here
        that cleanup the test environment after the test() methods in this class have been executed.
        This method does not have access to the class `self` reference.  This method
        is able to call static methods within this same class.
        """
        # FreeCAD.Console.PrintMessage("TestPathAdaptive.tearDownClass()\n")

        # Close geometry document without saving
        if not cls.needsInit:
            FreeCAD.closeDocument(cls.doc.Name)

    # Setup and tear down methods called before and after each unit test
    def setUp(self):
        """setUp()...
        This method is called prior to each `test()` method.  Add code and objects here
        that are needed for multiple `test()` methods.
        """
        if self.needsInit:
            self.initClass()

    def tearDown(self):
        """tearDown()...
        This method is called after each test() method. Add cleanup instructions here.
        Such cleanup instructions will likely undo those in the setUp() method.
        """
        pass

    # Unit tests
    def test00(self):
        """test00() Empty test."""
        return

    def test01(self):
        """test01() Verify path generated on Face18, outside, with tool compensation."""

        # Instantiate a Profile operation and set Base Geometry
        profile = PathProfile.Create("Profile1")
        profile.Base = [(self.doc.Body, ["Face18"])]  # (base, subs_list)
        profile.Label = "test01+"
        profile.Comment = (
            "test01() Verify path generated on Face18, outside, with tool compensation."
        )

        # Set additional operation properties
        # setDepthsAndHeights(adaptive)
        profile.processCircles = True
        profile.processHoles = True
        profile.UseComp = True
        profile.Direction = "CW"
        _addViewProvider(profile)
        self.doc.recompute()

        moves = getGcodeMoves(profile.Path.Commands, includeRapids=False)
        operationMoves = ";  ".join(moves)
        # FreeCAD.Console.PrintMessage("test01_moves: " + operationMoves + "\n")

        expected_moves = (
            "G1 X16.47 Y16.47 Z10.0;  G3 I-2.48 J-2.48 K0.0 X13.93 Y17.5 Z10.0;  "
            "G1 X-13.93 Y17.5 Z10.0;  G3 I-0.06 J-3.51 K0.0 X-17.5 Y13.93 Z10.0;  "
            "G1 X-17.5 Y-13.93 Z10.0;  G3 I3.51 J-0.06 K0.0 X-13.93 Y-17.5 Z10.0;  "
            "G1 X13.93 Y-17.5 Z10.0;  G3 I0.06 J3.51 K0.0 X17.5 Y-13.93 Z10.0;  "
            "G1 X17.5 Y13.93 Z10.0;  G3 I-3.51 J0.06 K0.0 X16.47 Y16.47 Z10.0;  "
            "G1 X23.54 Y23.54 Z10.0;  G2 I-9.55 J-9.55 K0.0 X27.5 Y14.0 Z10.0;  "
            "G1 X27.5 Y-14.0 Z10.0;  G2 I-13.5 J0.0 K0.0 X14.0 Y-27.5 Z10.0;  "
            "G1 X-14.0 Y-27.5 Z10.0;  G2 I0.0 J13.5 K0.0 X-27.5 Y-14.0 Z10.0;  "
            "G1 X-27.5 Y14.0 Z10.0;  G2 I13.5 J-0.0 K0.0 X-14.0 Y27.5 Z10.0;  "
            "G1 X14.0 Y27.5 Z10.0;  G2 I-0.0 J-13.5 K0.0 X23.54 Y23.54 Z10.0"
        )
        self.assertTrue(
            expected_moves == operationMoves,
            "expected_moves: {}\noperationMoves: {}".format(expected_moves, operationMoves),
        )

    def test02(self):
        """test02() Verify path generated on Face18, outside, without compensation."""

        # Instantiate a Profile operation and set Base Geometry
        profile = PathProfile.Create("Profile2")
        profile.Base = [(self.doc.Body, ["Face18"])]  # (base, subs_list)
        profile.Label = "test02+"
        profile.Comment = "test02() Verify path generated on Face18, outside, without compensation."

        # Set additional operation properties
        # setDepthsAndHeights(adaptive)
        profile.processCircles = True
        profile.processHoles = True
        profile.UseComp = False
        profile.Direction = "CW"
        _addViewProvider(profile)
        self.doc.recompute()

        moves = getGcodeMoves(profile.Path.Commands, includeRapids=False)
        operationMoves = ";  ".join(moves)
        # FreeCAD.Console.PrintMessage("test02_moves: " + operationMoves + "\n")

        expected_moves = (
            "G1 X18.24 Y18.24 Z10.0;  G3 I-4.24 J-4.24 K0.0 X14.0 Y20.0 Z10.0;  "
            "G1 X-14.0 Y20.0 Z10.0;  G3 I0.0 J-6.0 K0.0 X-20.0 Y14.0 Z10.0;  "
            "G1 X-20.0 Y-14.0 Z10.0;  G3 I6.0 J0.0 K0.0 X-14.0 Y-20.0 Z10.0;  "
            "G1 X14.0 Y-20.0 Z10.0;  G3 I-0.0 J6.0 K0.0 X20.0 Y-14.0 Z10.0;  "
            "G1 X20.0 Y14.0 Z10.0;  G3 I-6.0 J-0.0 K0.0 X18.24 Y18.24 Z10.0;  "
            "G1 X21.78 Y21.78 Z10.0;  G2 I-7.78 J-7.78 K0.0 X25.0 Y14.0 Z10.0;  "
            "G1 X25.0 Y-14.0 Z10.0;  G2 I-11.0 J0.0 K0.0 X14.0 Y-25.0 Z10.0;  "
            "G1 X-14.0 Y-25.0 Z10.0;  G2 I0.0 J11.0 K0.0 X-25.0 Y-14.0 Z10.0;  "
            "G1 X-25.0 Y14.0 Z10.0;  G2 I11.0 J-0.0 K0.0 X-14.0 Y25.0 Z10.0;  "
            "G1 X14.0 Y25.0 Z10.0;  G2 I-0.0 J-11.0 K0.0 X21.78 Y21.78 Z10.0"
        )

        self.assertTrue(
            expected_moves == operationMoves,
            "expected_moves: {}\noperationMoves: {}".format(expected_moves, operationMoves),
        )

    def test03(self):
        """test03() Verify path generated on Face18, outside,
        with compensation and extra offset -radius."""

        # Instantiate a Profile operation and set Base Geometry
        profile = PathProfile.Create("Profile3")
        profile.Base = [(self.doc.Body, ["Face18"])]  # (base, subs_list)
        profile.Label = "test03+"
        profile.Comment = (
            "test03() Verify path generated on Face4, " "with compensation and extra offset -radius"
        )

        # Set additional operation properties
        # setDepthsAndHeights(adaptive)
        profile.processCircles = True
        profile.processHoles = True
        profile.UseComp = True
        profile.Direction = "CW"
        profile.OffsetExtra = -profile.OpToolDiameter / 2.0
        _addViewProvider(profile)
        self.doc.recompute()

        moves = getGcodeMoves(profile.Path.Commands, includeRapids=False)
        operationMoves = ";  ".join(moves)
        # FreeCAD.Console.PrintMessage("test03_moves: " + operationMoves + "\n")

        expected_moves = (
            "G1 X18.24 Y18.24 Z10.0;  G3 I-4.24 J-4.24 K0.0 X14.0 Y20.0 Z10.0;  "
            "G1 X-14.0 Y20.0 Z10.0;  G3 I0.0 J-6.0 K0.0 X-20.0 Y14.0 Z10.0;  "
            "G1 X-20.0 Y-14.0 Z10.0;  G3 I6.0 J0.0 K0.0 X-14.0 Y-20.0 Z10.0;  "
            "G1 X14.0 Y-20.0 Z10.0;  G3 I-0.0 J6.0 K0.0 X20.0 Y-14.0 Z10.0;  "
            "G1 X20.0 Y14.0 Z10.0;  G3 I-6.0 J-0.0 K0.0 X18.24 Y18.24 Z10.0;  "
            "G1 X21.78 Y21.78 Z10.0;  G2 I-7.78 J-7.78 K0.0 X25.0 Y14.0 Z10.0;  "
            "G1 X25.0 Y-14.0 Z10.0;  G2 I-11.0 J0.0 K0.0 X14.0 Y-25.0 Z10.0;  "
            "G1 X-14.0 Y-25.0 Z10.0;  G2 I0.0 J11.0 K0.0 X-25.0 Y-14.0 Z10.0;  "
            "G1 X-25.0 Y14.0 Z10.0;  G2 I11.0 J-0.0 K0.0 X-14.0 Y25.0 Z10.0;  "
            "G1 X14.0 Y25.0 Z10.0;  G2 I-0.0 J-11.0 K0.0 X21.78 Y21.78 Z10.0"
        )

        self.assertTrue(
            expected_moves == operationMoves,
            "expected_moves: {}\noperationMoves: {}".format(expected_moves, operationMoves),
        )


def _addViewProvider(profileOp):
    if FreeCAD.GuiUp:
        PathOpGui = PathProfileGui.PathOpGui
        cmdRes = PathProfileGui.Command.res
        profileOp.ViewObject.Proxy = PathOpGui.ViewProvider(profileOp.ViewObject, cmdRes)



# ***************************************************************************
# *   Copyright (c) 2023 Syres                                              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the TechDraw FillTemplateFields GuiCommand."""

__title__ = "TechDrawTools.CommandFillTemplateFields"
__author__ = "Syres"
__url__ = "https://www.freecad.org"
__version__ = "00.02"
__date__ = "2023/12/07"

from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import FreeCADGui as Gui
import codecs
import csv
import os.path

import TechDrawTools


class CommandFillTemplateFields:
    """Use document info to populate the template fields."""

    def GetResources(self):
        """Return a dictionary with data that will be used by the button or menu item."""
        return {
            "Pixmap": "actions/TechDraw_FillTemplateFields.svg",
            "Accel": "",
            "MenuText": QT_TRANSLATE_NOOP(
                "TechDraw_FillTemplateFields", "Update Template Fields"
            ),
            "ToolTip": QT_TRANSLATE_NOOP(
                "TechDraw_FillTemplateFields",
                "Uses document info to populate the template fields",
            ),
        }

    def Activated(self):
        """Run the following code when the command is activated (button press)."""
        TechDrawTools.TaskFillTemplateFields()

    def IsActive(self):
        """Return True when the command should be active
        or False when it should be disabled (greyed)."""
        if App.ActiveDocument:
            objs = App.ActiveDocument.findObjects(Type="TechDraw::DrawPage")
            if not objs:
                return false

            for obj in objs:
                file_path = (
                    App.getResourceDir()
                    + "Mod/TechDraw/CSVdata/FillTemplateFields.csv"
                )
                if os.path.exists(file_path):
                    listofkeys = [
                        "CreatedByChkLst",
                        "ScaleChkLst",
                        "LabelChkLst",
                        "CommentChkLst",
                        "CompanyChkLst",
                        "LicenseChkLst",
                        "CreatedDateChkLst",
                        "LastModifiedDateChkLst",
                    ]
                    with codecs.open(file_path, encoding="utf-8") as fp:
                        reader = csv.DictReader(fp)
                        page = obj
                        texts = page.Template.EditableTexts
                        if (
                            texts
                            and os.path.exists(file_path)
                            and listofkeys == reader.fieldnames
                            and obj.Views != []
                        ):
                            return True
            return false



#
# The command must be "registered" with a unique name by calling its class.
Gui.addCommand("TechDraw_FillTemplateFields", CommandFillTemplateFields())




# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *   Copyright (c) 2020 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

from PySide import QtCore, QtGui
import FreeCAD
import FreeCADGui
import Path
import Path.Base.Gui.IconViewProvider as PathIconViewProvider
import Path.Base.Gui.PropertyEditor as PathPropertyEditor
import Path.Base.PropertyBag as PathPropertyBag
import Path.Base.Util as PathUtil
import re


__title__ = "Property Bag Editor"
__author__ = "sliptonic (Brad Collette)"
__url__ = "https://www.freecad.org"
__doc__ = "Task panel editor for a PropertyBag"

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())

translate = FreeCAD.Qt.translate


class ViewProvider(object):
    """ViewProvider for a PropertyBag.
    It's sole job is to provide an icon and invoke the TaskPanel on edit."""

    def __init__(self, vobj, name):
        Path.Log.track(name)
        vobj.Proxy = self
        self.icon = name
        # mode = 2
        self.obj = None
        self.vobj = None

    def attach(self, vobj):
        Path.Log.track()
        self.vobj = vobj
        self.obj = vobj.Object

    def getIcon(self):
        return ":/icons/Path-SetupSheet.svg"

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def getDisplayMode(self, mode):
        return "Default"

    def setEdit(self, vobj, mode=0):
        Path.Log.track()
        taskPanel = TaskPanel(vobj)
        FreeCADGui.Control.closeDialog()
        FreeCADGui.Control.showDialog(taskPanel)
        taskPanel.setupUi()
        return True

    def unsetEdit(self, vobj, mode):
        FreeCADGui.Control.closeDialog()
        return

    def claimChildren(self):
        return []

    def doubleClicked(self, vobj):
        self.setEdit(vobj)


class Delegate(QtGui.QStyledItemDelegate):
    RoleObject = QtCore.Qt.UserRole + 1
    RoleProperty = QtCore.Qt.UserRole + 2
    RoleEditor = QtCore.Qt.UserRole + 3

    # def paint(self, painter, option, index):
    #    #Path.Log.track(index.column(), type(option))

    def createEditor(self, parent, option, index):
        editor = PathPropertyEditor.Editor(
            index.data(self.RoleObject), index.data(self.RoleProperty)
        )
        index.model().setData(index, editor, self.RoleEditor)
        return editor.widget(parent)

    def setEditorData(self, widget, index):
        Path.Log.track(index.row(), index.column())
        index.data(self.RoleEditor).setEditorData(widget)

    def setModelData(self, widget, model, index):
        Path.Log.track(index.row(), index.column())
        editor = index.data(self.RoleEditor)
        editor.setModelData(widget)
        index.model().setData(index, editor.displayString(), QtCore.Qt.DisplayRole)

    def updateEditorGeometry(self, widget, option, index):
        widget.setGeometry(option.rect)


class PropertyCreate(object):
    def __init__(self, obj, grp, typ, another):
        self.obj = obj
        self.form = FreeCADGui.PySideUic.loadUi(":panels/PropertyCreate.ui")

        obj.Proxy.refreshCustomPropertyGroups()
        groups = obj.CustomPropertyGroups
        if not isinstance(groups, (list, tuple)):
            groups = [groups]
        for g in sorted(groups):
            self.form.propertyGroup.addItem(g)
        if grp:
            self.form.propertyGroup.setCurrentText(grp)

        for t in sorted(PathPropertyBag.SupportedPropertyType):
            self.form.propertyType.addItem(t)
            if PathPropertyBag.SupportedPropertyType[t] == typ:
                typ = t
        if typ:
            self.form.propertyType.setCurrentText(typ)
        else:
            self.form.propertyType.setCurrentText("String")
        self.form.createAnother.setChecked(another)

        self.form.propertyGroup.currentTextChanged.connect(self.updateUI)
        self.form.propertyGroup.currentIndexChanged.connect(self.updateUI)
        self.form.propertyName.textChanged.connect(self.updateUI)
        self.form.propertyType.currentIndexChanged.connect(self.updateUI)
        self.form.propertyEnum.textChanged.connect(self.updateUI)

    def updateUI(self):
        typeSet = True
        if self.propertyIsEnumeration():
            self.form.labelEnum.setEnabled(True)
            self.form.propertyEnum.setEnabled(True)
            typeSet = self.form.propertyEnum.text().strip() != ""
        else:
            self.form.labelEnum.setEnabled(False)
            self.form.propertyEnum.setEnabled(False)
            if self.form.propertyEnum.text().strip():
                self.form.propertyEnum.setText("")

        ok = self.form.buttonBox.button(QtGui.QDialogButtonBox.Ok)

        if not re.match("^[A-Za-z0-9_]*$", self.form.propertyName.text()):
            typeSet = False

        if typeSet and self.propertyGroup():
            ok.setEnabled(True)
        else:
            ok.setEnabled(False)

    def propertyName(self):
        return self.form.propertyName.text().strip()

    def propertyGroup(self):
        return self.form.propertyGroup.currentText().strip()

    def propertyType(self):
        return PathPropertyBag.SupportedPropertyType[self.form.propertyType.currentText()].strip()

    def propertyInfo(self):
        return self.form.propertyInfo.toPlainText().strip()

    def createAnother(self):
        return self.form.createAnother.isChecked()

    def propertyEnumerations(self):
        return [s.strip() for s in self.form.propertyEnum.text().strip().split(",")]

    def propertyIsEnumeration(self):
        return self.propertyType() == "App::PropertyEnumeration"

    def exec_(self, name):
        if name:
            # property exists - this is an edit operation
            self.form.propertyName.setText(name)
            if self.propertyIsEnumeration():
                self.form.propertyEnum.setText(",".join(self.obj.getEnumerationsOfProperty(name)))
            self.form.propertyInfo.setText(self.obj.getDocumentationOfProperty(name))

            self.form.labelName.setEnabled(False)
            self.form.propertyName.setEnabled(False)
            self.form.labelType.setEnabled(False)
            self.form.propertyType.setEnabled(False)
            self.form.createAnother.setEnabled(False)

        else:
            self.form.propertyName.setText("")
            self.form.propertyInfo.setText("")
            self.form.propertyEnum.setText("")
            # self.form.propertyName.setFocus()

        self.updateUI()

        return self.form.exec_()


class TaskPanel(object):
    ColumnName = 0
    # ColumnType = 1
    ColumnVal = 1
    # TableHeaders = ['Property', 'Type', 'Value']
    TableHeaders = ["Property", "Value"]

    def __init__(self, vobj):
        self.obj = vobj.Object
        self.props = sorted(self.obj.Proxy.getCustomProperties())
        self.form = FreeCADGui.PySideUic.loadUi(":panels/PropertyBag.ui")

        # initialized later
        self.model = None
        self.delegate = None
        FreeCAD.ActiveDocument.openTransaction("Edit PropertyBag")

    def updateData(self, topLeft, bottomRight):
        pass

    def _setupProperty(self, i, name):
        if name not in self.obj.PropertiesList:
            Path.Log.warning(f"Property '{name}' not found in object {self.obj.Name}")
            return
        prop_type_id = self.obj.getTypeIdOfProperty(name)
        try:
            typ = PathPropertyBag.getPropertyTypeName(prop_type_id)
        except IndexError:
            Path.Log.error(
                f"Unknown property type id '{prop_type_id}' for property '{name}' in object {self.obj.Name}"
            )
            return
        val = PathUtil.getPropertyValueString(self.obj, name)
        info = self.obj.getDocumentationOfProperty(name)

        self.model.setData(self.model.index(i, self.ColumnName), name, QtCore.Qt.EditRole)
        # self.model.setData(self.model.index(i, self.ColumnType), typ,       QtCore.Qt.EditRole)
        self.model.setData(self.model.index(i, self.ColumnVal), self.obj, Delegate.RoleObject)
        self.model.setData(self.model.index(i, self.ColumnVal), name, Delegate.RoleProperty)
        self.model.setData(self.model.index(i, self.ColumnVal), val, QtCore.Qt.DisplayRole)

        self.model.setData(self.model.index(i, self.ColumnName), typ, QtCore.Qt.ToolTipRole)
        # self.model.setData(self.model.index(i, self.ColumnType), info,      QtCore.Qt.ToolTipRole)
        self.model.setData(self.model.index(i, self.ColumnVal), info, QtCore.Qt.ToolTipRole)

        self.model.item(i, self.ColumnName).setEditable(False)
        # self.model.item(i, self.ColumnType).setEditable(False)

    def setupUi(self):
        Path.Log.track()

        self.delegate = Delegate(self.form)
        self.model = QtGui.QStandardItemModel(len(self.props), len(self.TableHeaders), self.form)
        self.model.setHorizontalHeaderLabels(self.TableHeaders)

        for i, name in enumerate(self.props):
            self._setupProperty(i, name)

        self.form.table.setModel(self.model)
        self.form.table.setItemDelegateForColumn(self.ColumnVal, self.delegate)
        self.form.table.resizeColumnsToContents()

        self.model.dataChanged.connect(self.updateData)
        self.form.table.selectionModel().selectionChanged.connect(self.propertySelected)
        self.form.add.clicked.connect(self.propertyAdd)
        self.form.remove.clicked.connect(self.propertyRemove)
        self.form.modify.clicked.connect(self.propertyModify)
        self.form.table.doubleClicked.connect(self.propertyModifyIndex)
        self.propertySelected([])

    def accept(self):
        FreeCAD.ActiveDocument.commitTransaction()
        FreeCADGui.ActiveDocument.resetEdit()
        FreeCADGui.Control.closeDialog()
        FreeCAD.ActiveDocument.recompute()

    def reject(self):
        FreeCAD.ActiveDocument.abortTransaction()
        FreeCADGui.Control.closeDialog()
        FreeCAD.ActiveDocument.recompute()

    def propertySelected(self, selection):
        Path.Log.track()
        if selection:
            self.form.modify.setEnabled(True)
            self.form.remove.setEnabled(True)
        else:
            self.form.modify.setEnabled(False)
            self.form.remove.setEnabled(False)

    def addCustomProperty(self, obj, dialog):
        name = dialog.propertyName()
        typ = dialog.propertyType()
        grp = dialog.propertyGroup()
        info = dialog.propertyInfo()
        if name:
            propname = self.obj.Proxy.addCustomProperty(typ, name, grp, info)
            if dialog.propertyIsEnumeration():
                setattr(self.obj, name, dialog.propertyEnumerations())
            return (propname, info)
        return (None, None)

    def propertyAdd(self):
        Path.Log.track()
        more = False
        grp = None
        typ = None
        while True:
            dialog = PropertyCreate(self.obj, grp, typ, more)
            if dialog.exec_(None):
                # if we block signals the view doesn't get updated, surprise, surprise
                # self.model.blockSignals(True)
                name, info = self.addCustomProperty(self.obj, dialog)
                if name:
                    index = 0
                    for i in range(self.model.rowCount()):
                        index = i
                        if (
                            self.model.item(i, self.ColumnName).data(QtCore.Qt.EditRole)
                            > dialog.propertyName()
                        ):
                            break
                    self.model.insertRows(index, 1)
                    self._setupProperty(index, name)
                    self.form.table.selectionModel().setCurrentIndex(
                        self.model.index(index, 0), QtCore.QItemSelectionModel.Rows
                    )
                    # self.model.blockSignals(False)
                    more = dialog.createAnother()
                else:
                    more = False
            else:
                more = False
            if not more:
                break

    def propertyModifyIndex(self, index):
        Path.Log.track(index.row(), index.column())
        row = index.row()

        obj = self.model.item(row, self.ColumnVal).data(Delegate.RoleObject)
        prop = self.model.item(row, self.ColumnVal).data(Delegate.RoleProperty)
        grp = obj.getGroupOfProperty(prop)
        typ = obj.getTypeIdOfProperty(prop)

        dialog = PropertyCreate(self.obj, grp, typ, False)
        if dialog.exec_(prop):
            val = getattr(obj, prop)
            obj.removeProperty(prop)
            name, info = self.addCustomProperty(self.obj, dialog)
            try:
                setattr(obj, prop, val)
            except Exception:
                # this can happen if the old enumeration value doesn't exist anymore
                pass
            newVal = PathUtil.getPropertyValueString(obj, prop)
            self.model.setData(self.model.index(row, self.ColumnVal), newVal, QtCore.Qt.DisplayRole)

            # self.model.setData(self.model.index(row, self.ColumnType), info, QtCore.Qt.ToolTipRole)
            self.model.setData(self.model.index(row, self.ColumnVal), info, QtCore.Qt.ToolTipRole)

    def propertyModify(self):
        Path.Log.track()
        rows = []
        for index in self.form.table.selectionModel().selectedIndexes():
            row = index.row()
            if row in rows:
                continue
            rows.append(row)

            self.propertyModifyIndex(index)

    def propertyRemove(self):
        Path.Log.track()
        # first find all rows which need to be removed
        rows = []
        for index in self.form.table.selectionModel().selectedIndexes():
            if not index.row() in rows:
                rows.append(index.row())

        # then remove them in reverse order so the indexes of the remaining rows
        # to delete are still valid
        for row in reversed(sorted(rows)):
            self.obj.removeProperty(self.model.item(row).data(QtCore.Qt.EditRole))
            self.model.removeRow(row)


def Create(name="PropertyBag"):
    """Create(name = 'PropertyBag') ... creates a new setup sheet"""
    FreeCAD.ActiveDocument.openTransaction("Create PropertyBag")
    pcont = PathPropertyBag.Create(name)
    PathIconViewProvider.Attach(pcont.ViewObject, name)
    return pcont


PathIconViewProvider.RegisterViewProvider("PropertyBag", ViewProvider)


class PropertyBagCreateCommand(object):
    """Command to create a property container object"""

    def __init__(self):
        pass

    def GetResources(self):
        return {
            "MenuText": translate("CAM_PropertyBag", "Property Bag"),
            "ToolTip": translate(
                "CAM_PropertyBag",
                "Creates an object which can be used to store reference properties",
            ),
        }

    def IsActive(self):
        return not FreeCAD.ActiveDocument is None

    def Activated(self):
        sel = FreeCADGui.Selection.getSelectionEx()
        obj = Create()
        body = None
        if sel:
            if "PartDesign::Body" == sel[0].Object.TypeId:
                body = sel[0].Object
            elif hasattr(sel[0].Object, "getParentGeoFeatureGroup"):
                body = sel[0].Object.getParentGeoFeatureGroup()
        if body:
            obj.Label = "Attributes"
            group = body.Group
            group.append(obj)
            body.Group = group


if FreeCAD.GuiUp:
    FreeCADGui.addCommand("CAM_PropertyBag", PropertyBagCreateCommand())

FreeCAD.Console.PrintLog("Loading PathPropertyBagGui… done\n")



# ***************************************************************************
# *   Copyright (c) 2018 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "Objects FEM unit tests"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

import sys
import unittest
from os.path import join

import FreeCAD

import ObjectsFem
from . import support_utils as testtools
from .support_utils import fcc_print


class TestObjectCreate(unittest.TestCase):
    fcc_print("import TestObjectCreate")

    # ********************************************************************************************
    def setUp(self):
        # setUp is executed before every test

        # new document
        self.document = FreeCAD.newDocument(self.__class__.__name__)

    # ********************************************************************************************
    def tearDown(self):
        # tearDown is executed after every test
        FreeCAD.closeDocument(self.document.Name)

    # ********************************************************************************************
    def test_00print(self):
        # since method name starts with 00 this will be run first
        # this test just prints a line with stars

        fcc_print(
            "\n{0}\n{1} run FEM TestObjectCreate tests {2}\n{0}".format(
                100 * "*", 10 * "*", 58 * "*"
            )
        )

    # ********************************************************************************************
    def test_femobjects_make(self):
        doc = create_all_fem_objects_doc(self.document)

        # count the def make in ObjectsFem module
        # if FEM VTK post processing is disabled, we are not able to create VTK post objects
        vtk_objects_used = "BUILD_FEM_VTK" in FreeCAD.__cmake__
        count_defmake = testtools.get_defmake_count(vtk_objects_used)

        # TODO if the children are added to the analysis, they show up twice on Tree
        # thus they are not added to the analysis group ATM
        # https://forum.freecad.org/viewtopic.php?t=25283
        # thus they should not be counted
        # solver children: equations --> 10
        # gmsh mesh children: group, region, boundary layer --> 3
        # result children: mesh result --> 1
        # analysis itself is not in analysis group --> 1
        # vtk post pipeline children: region, scalar, cut, wrap, contour --> 5
        # vtk python post objects: glyph, 6x data extraction --> 7

        subtraction = 15
        if vtk_objects_used:
            subtraction += 12
            if not ("BUILD_FEM_VTK_PYTHON" in FreeCAD.__cmake__):
                # remove the 3 data visualization objects that would be in the Analysis
                # if they would be available (Lineplot, histogram, table)
                subtraction += 3

        self.assertEqual(len(doc.Analysis.Group), count_defmake - subtraction)

        # if vtk used, but python API is not available, the vtk python based objects "def make" functions
        # have been counted, but will not be executed to create objects
        failed = 0
        if vtk_objects_used and not ("BUILD_FEM_VTK_PYTHON" in FreeCAD.__cmake__):
            # the 7 objects also counted in subtraction, +3 additional objects that are
            # added directly to the analysis
            failed += 10

        self.assertEqual(len(doc.Objects), count_defmake - failed)

        fcc_print(
            "doc objects count: {}, method: {}".format(
                len(doc.Objects), sys._getframe().f_code.co_name
            )
        )

        # save the file
        save_fc_file = join(
            testtools.get_fem_test_tmp_dir("objects_create_all"), "all_objects.FCStd"
        )
        fcc_print(f"Save FreeCAD all objects file to {save_fc_file} ...")
        self.document.saveAs(save_fc_file)


# ************************************************************************************************
# ************************************************************************************************
class TestObjectType(unittest.TestCase):
    fcc_print("import TestObjectType")

    # ********************************************************************************************
    def setUp(self):
        # setUp is executed before every test

        # new document
        self.document = FreeCAD.newDocument(self.__class__.__name__)

    # ********************************************************************************************
    def tearDown(self):
        # tearDown is executed after every test
        FreeCAD.closeDocument(self.document.Name)

    # ********************************************************************************************
    def test_00print(self):
        # since method name starts with 00 this will be run first
        # this test just prints a line with stars

        fcc_print(
            "\n{0}\n{1} run FEM TestObjectType tests {2}\n{0}".format(100 * "*", 10 * "*", 60 * "*")
        )

    # ********************************************************************************************
    def test_femobjects_type(self):
        doc = self.document
        create_all_fem_objects_doc

        from femtools.femutils import type_of_obj

        self.assertEqual("Fem::FemAnalysis", type_of_obj(ObjectsFem.makeAnalysis(doc)))
        self.assertEqual(
            "Fem::ConstantVacuumPermittivity",
            type_of_obj(ObjectsFem.makeConstantVacuumPermittivity(doc)),
        )
        self.assertEqual(
            "Fem::ConstraintBearing", type_of_obj(ObjectsFem.makeConstraintBearing(doc))
        )
        self.assertEqual(
            "Fem::ConstraintBodyHeatSource",
            type_of_obj(ObjectsFem.makeConstraintBodyHeatSource(doc)),
        )
        self.assertEqual(
            "Fem::ConstraintContact", type_of_obj(ObjectsFem.makeConstraintContact(doc))
        )
        self.assertEqual(
            "Fem::ConstraintCurrentDensity",
            type_of_obj(ObjectsFem.makeConstraintCurrentDensity(doc)),
        )
        self.assertEqual(
            "Fem::ConstraintDisplacement", type_of_obj(ObjectsFem.makeConstraintDisplacement(doc))
        )
        self.assertEqual(
            "Fem::ConstraintElectrostaticPotential",
            type_of_obj(ObjectsFem.makeConstraintElectrostaticPotential(doc)),
        )
        self.assertEqual(
            "Fem::ConstraintElectricChargeDensity",
            type_of_obj(ObjectsFem.makeConstraintElectricChargeDensity(doc)),
        )
        self.assertEqual("Fem::ConstraintFixed", type_of_obj(ObjectsFem.makeConstraintFixed(doc)))
        self.assertEqual(
            "Fem::ConstraintRigidBody", type_of_obj(ObjectsFem.makeConstraintRigidBody(doc))
        )
        self.assertEqual(
            "Fem::ConstraintFlowVelocity", type_of_obj(ObjectsFem.makeConstraintFlowVelocity(doc))
        )
        self.assertEqual(
            "Fem::ConstraintFluidBoundary", type_of_obj(ObjectsFem.makeConstraintFluidBoundary(doc))
        )
        self.assertEqual("Fem::ConstraintSpring", type_of_obj(ObjectsFem.makeConstraintSpring(doc)))
        self.assertEqual("Fem::ConstraintForce", type_of_obj(ObjectsFem.makeConstraintForce(doc)))
        self.assertEqual("Fem::ConstraintGear", type_of_obj(ObjectsFem.makeConstraintGear(doc)))
        self.assertEqual(
            "Fem::ConstraintHeatflux", type_of_obj(ObjectsFem.makeConstraintHeatflux(doc))
        )
        self.assertEqual(
            "Fem::ConstraintInitialFlowVelocity",
            type_of_obj(ObjectsFem.makeConstraintInitialFlowVelocity(doc)),
        )
        self.assertEqual(
            "Fem::ConstraintInitialPressure",
            type_of_obj(ObjectsFem.makeConstraintInitialPressure(doc)),
        )
        self.assertEqual(
            "Fem::ConstraintInitialTemperature",
            type_of_obj(ObjectsFem.makeConstraintInitialTemperature(doc)),
        )
        self.assertEqual(
            "Fem::ConstraintMagnetization", type_of_obj(ObjectsFem.makeConstraintMagnetization(doc))
        )
        self.assertEqual(
            "Fem::ConstraintPlaneRotation", type_of_obj(ObjectsFem.makeConstraintPlaneRotation(doc))
        )
        self.assertEqual(
            "Fem::ConstraintPressure", type_of_obj(ObjectsFem.makeConstraintPressure(doc))
        )
        self.assertEqual("Fem::ConstraintPulley", type_of_obj(ObjectsFem.makeConstraintPulley(doc)))
        self.assertEqual(
            "Fem::ConstraintSectionPrint", type_of_obj(ObjectsFem.makeConstraintSectionPrint(doc))
        )
        self.assertEqual(
            "Fem::ConstraintSelfWeight", type_of_obj(ObjectsFem.makeConstraintSelfWeight(doc))
        )
        self.assertEqual(
            "Fem::ConstraintCentrif", type_of_obj(ObjectsFem.makeConstraintCentrif(doc))
        )
        self.assertEqual(
            "Fem::ConstraintTemperature", type_of_obj(ObjectsFem.makeConstraintTemperature(doc))
        )
        self.assertEqual("Fem::ConstraintTie", type_of_obj(ObjectsFem.makeConstraintTie(doc)))
        self.assertEqual(
            "Fem::ConstraintTransform", type_of_obj(ObjectsFem.makeConstraintTransform(doc))
        )
        self.assertEqual("Fem::ElementFluid1D", type_of_obj(ObjectsFem.makeElementFluid1D(doc)))
        self.assertEqual(
            "Fem::ElementGeometry1D", type_of_obj(ObjectsFem.makeElementGeometry1D(doc))
        )
        self.assertEqual(
            "Fem::ElementGeometry2D", type_of_obj(ObjectsFem.makeElementGeometry2D(doc))
        )
        self.assertEqual(
            "Fem::ElementRotation1D", type_of_obj(ObjectsFem.makeElementRotation1D(doc))
        )
        materialsolid = ObjectsFem.makeMaterialSolid(doc)
        self.assertEqual("Fem::MaterialCommon", type_of_obj(ObjectsFem.makeMaterialFluid(doc)))
        self.assertEqual("Fem::MaterialCommon", type_of_obj(materialsolid))
        self.assertEqual(
            "Fem::MaterialMechanicalNonlinear",
            type_of_obj(ObjectsFem.makeMaterialMechanicalNonlinear(doc, materialsolid)),
        )
        self.assertEqual(
            "Fem::MaterialReinforced", type_of_obj(ObjectsFem.makeMaterialReinforced(doc))
        )
        mesh = ObjectsFem.makeMeshGmsh(doc)
        self.assertEqual("Fem::FemMeshGmsh", type_of_obj(mesh))
        self.assertEqual(
            "Fem::MeshBoundaryLayer", type_of_obj(ObjectsFem.makeMeshBoundaryLayer(doc, mesh))
        )
        self.assertEqual("Fem::MeshGroup", type_of_obj(ObjectsFem.makeMeshGroup(doc, mesh)))
        self.assertEqual("Fem::MeshRegion", type_of_obj(ObjectsFem.makeMeshRegion(doc, mesh)))
        self.assertEqual("Fem::FemMeshNetgen", type_of_obj(ObjectsFem.makeMeshNetgen(doc)))
        self.assertEqual(
            "Fem::FemMeshShapeNetgenObject", type_of_obj(ObjectsFem.makeMeshNetgenLegacy(doc))
        )
        self.assertEqual("Fem::MeshResult", type_of_obj(ObjectsFem.makeMeshResult(doc)))
        self.assertEqual("Fem::ResultMechanical", type_of_obj(ObjectsFem.makeResultMechanical(doc)))
        solverelmer = ObjectsFem.makeSolverElmer(doc)
        self.assertEqual(
            "Fem::SolverCcxTools", type_of_obj(ObjectsFem.makeSolverCalculiXCcxTools(doc))
        )
        self.assertEqual("Fem::SolverCalculiX", type_of_obj(ObjectsFem.makeSolverCalculiX(doc)))
        self.assertEqual("Fem::SolverElmer", type_of_obj(solverelmer))
        self.assertEqual("Fem::SolverMystran", type_of_obj(ObjectsFem.makeSolverMystran(doc)))
        self.assertEqual("Fem::SolverZ88", type_of_obj(ObjectsFem.makeSolverZ88(doc)))
        self.assertEqual(
            "Fem::EquationElmerDeformation",
            type_of_obj(ObjectsFem.makeEquationDeformation(doc, solverelmer)),
        )
        self.assertEqual(
            "Fem::EquationElmerElasticity",
            type_of_obj(ObjectsFem.makeEquationElasticity(doc, solverelmer)),
        )
        self.assertEqual(
            "Fem::EquationElmerElectricforce",
            type_of_obj(ObjectsFem.makeEquationElectricforce(doc, solverelmer)),
        )
        self.assertEqual(
            "Fem::EquationElmerElectrostatic",
            type_of_obj(ObjectsFem.makeEquationElectrostatic(doc, solverelmer)),
        )
        self.assertEqual(
            "Fem::EquationElmerFlow", type_of_obj(ObjectsFem.makeEquationFlow(doc, solverelmer))
        )
        self.assertEqual(
            "Fem::EquationElmerFlux", type_of_obj(ObjectsFem.makeEquationFlux(doc, solverelmer))
        )
        self.assertEqual(
            "Fem::EquationElmerHeat", type_of_obj(ObjectsFem.makeEquationHeat(doc, solverelmer))
        )
        self.assertEqual(
            "Fem::EquationElmerMagnetodynamic2D",
            type_of_obj(ObjectsFem.makeEquationMagnetodynamic2D(doc, solverelmer)),
        )
        self.assertEqual(
            "Fem::EquationElmerMagnetodynamic",
            type_of_obj(ObjectsFem.makeEquationMagnetodynamic(doc, solverelmer)),
        )
        self.assertEqual(
            "Fem::EquationElmerStaticCurrent",
            type_of_obj(ObjectsFem.makeEquationStaticCurrent(doc, solverelmer)),
        )

        fcc_print(
            "doc objects count: {}, method: {}".format(
                len(doc.Objects), sys._getframe().f_code.co_name
            )
        )
        # TODO: vtk post objs, thus 5 obj less than test_femobjects_make
        self.assertEqual(len(doc.Objects), testtools.get_defmake_count(False))
        # TODO: use different type for material fluid and material solid

    # ********************************************************************************************
    def test_femobjects_isoftype(self):
        doc = self.document

        from femtools.femutils import is_of_type

        self.assertTrue(is_of_type(ObjectsFem.makeAnalysis(doc), "Fem::FemAnalysis"))
        self.assertTrue(
            is_of_type(
                ObjectsFem.makeConstantVacuumPermittivity(doc), "Fem::ConstantVacuumPermittivity"
            )
        )
        self.assertTrue(is_of_type(ObjectsFem.makeConstraintBearing(doc), "Fem::ConstraintBearing"))
        self.assertTrue(
            is_of_type(
                ObjectsFem.makeConstraintBodyHeatSource(doc), "Fem::ConstraintBodyHeatSource"
            )
        )
        self.assertTrue(is_of_type(ObjectsFem.makeConstraintContact(doc), "Fem::ConstraintContact"))
        self.assertTrue(
            is_of_type(
                ObjectsFem.makeConstraintCurrentDensity(doc), "Fem::ConstraintCurrentDensity"
            )
        )
        self.assertTrue(
            is_of_type(ObjectsFem.makeConstraintDisplacement(doc), "Fem::ConstraintDisplacement")
        )
        self.assertTrue(
            is_of_type(
                ObjectsFem.makeConstraintElectrostaticPotential(doc),
                "Fem::ConstraintElectrostaticPotential",
            )
        )
        self.assertTrue(
            is_of_type(
                ObjectsFem.makeConstraintElectricChargeDensity(doc),
                "Fem::ConstraintElectricChargeDensity",
            )
        )
        self.assertTrue(is_of_type(ObjectsFem.makeConstraintFixed(doc), "Fem::ConstraintFixed"))
        self.assertTrue(
            is_of_type(ObjectsFem.makeConstraintRigidBody(doc), "Fem::ConstraintRigidBody")
        )
        self.assertTrue(
            is_of_type(ObjectsFem.makeConstraintFlowVelocity(doc), "Fem::ConstraintFlowVelocity")
        )
        self.assertTrue(
            is_of_type(ObjectsFem.makeConstraintFluidBoundary(doc), "Fem::ConstraintFluidBoundary")
        )
        self.assertTrue(
            is_of_type(ObjectsFem.makeConstraintMagnetization(doc), "Fem::ConstraintMagnetization")
        )
        self.assertTrue(is_of_type(ObjectsFem.makeConstraintSpring(doc), "Fem::ConstraintSpring"))
        self.assertTrue(is_of_type(ObjectsFem.makeConstraintForce(doc), "Fem::ConstraintForce"))
        self.assertTrue(is_of_type(ObjectsFem.makeConstraintGear(doc), "Fem::ConstraintGear"))
        self.assertTrue(
            is_of_type(ObjectsFem.makeConstraintHeatflux(doc), "Fem::ConstraintHeatflux")
        )
        self.assertTrue(
            is_of_type(
                ObjectsFem.makeConstraintInitialFlowVelocity(doc),
                "Fem::ConstraintInitialFlowVelocity",
            )
        )
        self.assertTrue(
            is_of_type(
                ObjectsFem.makeConstraintInitialPressure(doc), "Fem::ConstraintInitialPressure"
            )
        )
        self.assertTrue(
            is_of_type(
                ObjectsFem.makeConstraintInitialTemperature(doc),
                "Fem::ConstraintInitialTemperature",
            )
        )
        self.assertTrue(
            is_of_type(ObjectsFem.makeConstraintPlaneRotation(doc), "Fem::ConstraintPlaneRotation")
        )
        self.assertTrue(
            is_of_type(ObjectsFem.makeConstraintPressure(doc), "Fem::ConstraintPressure")
        )
        self.assertTrue(is_of_type(ObjectsFem.makeConstraintPulley(doc), "Fem::ConstraintPulley"))
        self.assertTrue(
            is_of_type(ObjectsFem.makeConstraintSectionPrint(doc), "Fem::ConstraintSectionPrint")
        )
        self.assertTrue(
            is_of_type(ObjectsFem.makeConstraintSelfWeight(doc), "Fem::ConstraintSelfWeight")
        )
        self.assertTrue(is_of_type(ObjectsFem.makeConstraintCentrif(doc), "Fem::ConstraintCentrif"))
        self.assertTrue(
            is_of_type(ObjectsFem.makeConstraintTemperature(doc), "Fem::ConstraintTemperature")
        )
        self.assertTrue(is_of_type(ObjectsFem.makeConstraintTie(doc), "Fem::ConstraintTie"))
        self.assertTrue(
            is_of_type(ObjectsFem.makeConstraintTransform(doc), "Fem::ConstraintTransform")
        )
        self.assertTrue(is_of_type(ObjectsFem.makeElementFluid1D(doc), "Fem::ElementFluid1D"))
        self.assertTrue(is_of_type(ObjectsFem.makeElementGeometry1D(doc), "Fem::ElementGeometry1D"))
        self.assertTrue(is_of_type(ObjectsFem.makeElementGeometry2D(doc), "Fem::ElementGeometry2D"))
        self.assertTrue(is_of_type(ObjectsFem.makeElementRotation1D(doc), "Fem::ElementRotation1D"))
        materialsolid = ObjectsFem.makeMaterialSolid(doc)
        self.assertTrue(is_of_type(ObjectsFem.makeMaterialFluid(doc), "Fem::MaterialCommon"))
        self.assertTrue(is_of_type(materialsolid, "Fem::MaterialCommon"))
        self.assertTrue(
            is_of_type(
                ObjectsFem.makeMaterialMechanicalNonlinear(doc, materialsolid),
                "Fem::MaterialMechanicalNonlinear",
            )
        )
        self.assertTrue(
            is_of_type(ObjectsFem.makeMaterialReinforced(doc), "Fem::MaterialReinforced")
        )
        mesh = ObjectsFem.makeMeshGmsh(doc)
        self.assertTrue(is_of_type(mesh, "Fem::FemMeshGmsh"))
        self.assertTrue(
            is_of_type(ObjectsFem.makeMeshBoundaryLayer(doc, mesh), "Fem::MeshBoundaryLayer")
        )
        self.assertTrue(is_of_type(ObjectsFem.makeMeshGroup(doc, mesh), "Fem::MeshGroup"))
        self.assertTrue(is_of_type(ObjectsFem.makeMeshRegion(doc, mesh), "Fem::MeshRegion"))
        self.assertTrue(is_of_type(ObjectsFem.makeMeshNetgen(doc), "Fem::FemMeshNetgen"))
        self.assertTrue(
            is_of_type(ObjectsFem.makeMeshNetgenLegacy(doc), "Fem::FemMeshShapeNetgenObject")
        )
        self.assertTrue(is_of_type(ObjectsFem.makeMeshResult(doc), "Fem::MeshResult"))
        self.assertTrue(is_of_type(ObjectsFem.makeResultMechanical(doc), "Fem::ResultMechanical"))
        solverelmer = ObjectsFem.makeSolverElmer(doc)
        self.assertTrue(
            is_of_type(ObjectsFem.makeSolverCalculiXCcxTools(doc), "Fem::SolverCcxTools")
        )
        self.assertTrue(is_of_type(ObjectsFem.makeSolverCalculiX(doc), "Fem::SolverCalculiX"))
        self.assertTrue(is_of_type(solverelmer, "Fem::SolverElmer"))
        self.assertTrue(is_of_type(ObjectsFem.makeSolverMystran(doc), "Fem::SolverMystran"))
        self.assertTrue(is_of_type(ObjectsFem.makeSolverZ88(doc), "Fem::SolverZ88"))
        self.assertTrue(
            is_of_type(
                ObjectsFem.makeEquationDeformation(doc, solverelmer),
                "Fem::EquationElmerDeformation",
            )
        )
        self.assertTrue(
            is_of_type(
                ObjectsFem.makeEquationElasticity(doc, solverelmer), "Fem::EquationElmerElasticity"
            )
        )
        self.assertTrue(
            is_of_type(
                ObjectsFem.makeEquationElectricforce(doc, solverelmer),
                "Fem::EquationElmerElectricforce",
            )
        )
        self.assertTrue(
            is_of_type(
                ObjectsFem.makeEquationElectrostatic(doc, solverelmer),
                "Fem::EquationElmerElectrostatic",
            )
        )
        self.assertTrue(
            is_of_type(ObjectsFem.makeEquationFlow(doc, solverelmer), "Fem::EquationElmerFlow")
        )
        self.assertTrue(
            is_of_type(ObjectsFem.makeEquationFlux(doc, solverelmer), "Fem::EquationElmerFlux")
        )
        self.assertTrue(
            is_of_type(ObjectsFem.makeEquationHeat(doc, solverelmer), "Fem::EquationElmerHeat")
        )
        self.assertTrue(
            is_of_type(
                ObjectsFem.makeEquationMagnetodynamic2D(doc, solverelmer),
                "Fem::EquationElmerMagnetodynamic2D",
            )
        )
        self.assertTrue(
            is_of_type(
                ObjectsFem.makeEquationMagnetodynamic(doc, solverelmer),
                "Fem::EquationElmerMagnetodynamic",
            )
        )
        self.assertTrue(
            is_of_type(
                ObjectsFem.makeEquationStaticCurrent(doc, solverelmer),
                "Fem::EquationElmerStaticCurrent",
            )
        )

        fcc_print(
            "doc objects count: {}, method: {}".format(
                len(doc.Objects), sys._getframe().f_code.co_name
            )
        )
        # TODO: vtk post objs, thus 5 obj less than test_femobjects_make
        self.assertEqual(len(doc.Objects), testtools.get_defmake_count(False))

    # ********************************************************************************************
    def test_femobjects_derivedfromfem(self):
        # try to add all possible True types from inheritance chain see
        # https://forum.freecad.org/viewtopic.php?f=10&t=32625
        doc = self.document

        from femtools.femutils import is_derived_from

        # FemAnalysis
        analysis = ObjectsFem.makeAnalysis(doc)
        self.assertTrue(is_derived_from(analysis, "App::DocumentObject"))
        self.assertTrue(is_derived_from(analysis, "Fem::FemAnalysis"))

        # ConstantVacuumPermittivity
        constant_vacuumpermittivity = ObjectsFem.makeConstantVacuumPermittivity(doc)
        self.assertTrue(is_derived_from(constant_vacuumpermittivity, "App::DocumentObject"))
        self.assertTrue(is_derived_from(constant_vacuumpermittivity, "Fem::ConstraintPython"))
        self.assertTrue(
            is_derived_from(constant_vacuumpermittivity, "Fem::ConstantVacuumPermittivity")
        )

        # ConstraintBearing
        constraint_bearing = ObjectsFem.makeConstraintBearing(doc)
        self.assertTrue(is_derived_from(constraint_bearing, "App::DocumentObject"))
        self.assertTrue(is_derived_from(constraint_bearing, "Fem::Constraint"))
        self.assertTrue(is_derived_from(constraint_bearing, "Fem::ConstraintBearing"))

        # ConstraintBodyHeatSource
        constraint_body_heat_source = ObjectsFem.makeConstraintBodyHeatSource(doc)
        self.assertTrue(is_derived_from(constraint_body_heat_source, "App::DocumentObject"))
        self.assertTrue(is_derived_from(constraint_body_heat_source, "Fem::ConstraintPython"))
        self.assertTrue(
            is_derived_from(constraint_body_heat_source, "Fem::ConstraintBodyHeatSource")
        )

        # ConstraintContact
        constraint_contact = ObjectsFem.makeConstraintContact(doc)
        self.assertTrue(is_derived_from(constraint_contact, "App::DocumentObject"))
        self.assertTrue(is_derived_from(constraint_contact, "Fem::Constraint"))
        self.assertTrue(is_derived_from(constraint_contact, "Fem::ConstraintContact"))

        # ConstraintCurrentDensity
        constraint_currentdensity = ObjectsFem.makeConstraintCurrentDensity(doc)
        self.assertTrue(is_derived_from(constraint_currentdensity, "App::DocumentObject"))
        self.assertTrue(is_derived_from(constraint_currentdensity, "Fem::ConstraintPython"))
        self.assertTrue(is_derived_from(constraint_currentdensity, "Fem::ConstraintCurrentDensity"))

        # ConstraintDisplacement
        constraint_displacement = ObjectsFem.makeConstraintDisplacement(doc)
        self.assertTrue(is_derived_from(constraint_displacement, "App::DocumentObject"))
        self.assertTrue(is_derived_from(constraint_displacement, "Fem::Constraint"))
        self.assertTrue(is_derived_from(constraint_displacement, "Fem::ConstraintDisplacement"))

        # ConstraintElectrostaticPotential
        constraint_electorstatic_potential = ObjectsFem.makeConstraintElectrostaticPotential(doc)
        self.assertTrue(is_derived_from(constraint_electorstatic_potential, "App::DocumentObject"))
        self.assertTrue(
            is_derived_from(constraint_electorstatic_potential, "Fem::ConstraintPython")
        )
        self.assertTrue(
            is_derived_from(
                constraint_electorstatic_potential, "Fem::ConstraintElectrostaticPotential"
            )
        )

        # ConstraintElectricChargeDensity
        constraint_electric_charge_density = ObjectsFem.makeConstraintElectricChargeDensity(doc)
        self.assertTrue(is_derived_from(constraint_electric_charge_density, "App::DocumentObject"))
        self.assertTrue(
            is_derived_from(constraint_electric_charge_density, "Fem::ConstraintPython")
        )
        self.assertTrue(
            is_derived_from(
                constraint_electric_charge_density, "Fem::ConstraintElectricChargeDensity"
            )
        )

        # ConstraintFixed
        constraint_fixed = ObjectsFem.makeConstraintFixed(doc)
        self.assertTrue(is_derived_from(constraint_fixed, "App::DocumentObject"))
        self.assertTrue(is_derived_from(constraint_fixed, "Fem::Constraint"))
        self.assertTrue(is_derived_from(constraint_fixed, "Fem::ConstraintFixed"))

        # ConstraintRigidBody
        constraint_rigidbody = ObjectsFem.makeConstraintRigidBody(doc)
        self.assertTrue(is_derived_from(constraint_rigidbody, "App::DocumentObject"))
        self.assertTrue(is_derived_from(constraint_rigidbody, "Fem::Constraint"))
        self.assertTrue(is_derived_from(constraint_rigidbody, "Fem::ConstraintRigidBody"))

        # ConstraintFlowVelocity
        constraint_flow_velocity = ObjectsFem.makeConstraintFlowVelocity(doc)
        self.assertTrue(is_derived_from(constraint_flow_velocity, "App::DocumentObject"))
        self.assertTrue(is_derived_from(constraint_flow_velocity, "Fem::ConstraintPython"))
        self.assertTrue(is_derived_from(constraint_flow_velocity, "Fem::ConstraintFlowVelocity"))

        # ConstraintFluidBoundary
        constraint_fluid_boundary = ObjectsFem.makeConstraintFluidBoundary(doc)
        self.assertTrue(is_derived_from(constraint_fluid_boundary, "App::DocumentObject"))
        self.assertTrue(is_derived_from(constraint_fluid_boundary, "Fem::Constraint"))
        self.assertTrue(is_derived_from(constraint_fluid_boundary, "Fem::ConstraintFluidBoundary"))

        # ConstraintMagnetization
        constraint_magnetization = ObjectsFem.makeConstraintMagnetization(doc)
        self.assertTrue(is_derived_from(constraint_magnetization, "App::DocumentObject"))
        self.assertTrue(is_derived_from(constraint_magnetization, "Fem::ConstraintPython"))
        self.assertTrue(is_derived_from(constraint_magnetization, "Fem::ConstraintMagnetization"))

        # ConstraintSpring
        constraint_spring = ObjectsFem.makeConstraintSpring(doc)
        self.assertTrue(is_derived_from(constraint_spring, "App::DocumentObject"))
        self.assertTrue(is_derived_from(constraint_spring, "Fem::Constraint"))
        self.assertTrue(is_derived_from(constraint_spring, "Fem::ConstraintSpring"))

        # ConstraintForce
        constraint_force = ObjectsFem.makeConstraintForce(doc)
        self.assertTrue(is_derived_from(constraint_force, "App::DocumentObject"))
        self.assertTrue(is_derived_from(constraint_force, "Fem::Constraint"))
        self.assertTrue(is_derived_from(constraint_force, "Fem::ConstraintForce"))

        # ConstraintGear
        constraint_gear = ObjectsFem.makeConstraintGear(doc)
        self.assertTrue(is_derived_from(constraint_gear, "App::DocumentObject"))
        self.assertTrue(is_derived_from(constraint_gear, "Fem::Constraint"))
        self.assertTrue(is_derived_from(constraint_gear, "Fem::ConstraintGear"))

        # ConstraintHeatflux
        constraint_heat_flux = ObjectsFem.makeConstraintHeatflux(doc)
        self.assertTrue(is_derived_from(constraint_heat_flux, "App::DocumentObject"))
        self.assertTrue(is_derived_from(constraint_heat_flux, "Fem::Constraint"))
        self.assertTrue(is_derived_from(constraint_heat_flux, "Fem::ConstraintHeatflux"))

        # ConstraintInitialFlowVelocity
        constraint_initial_flow_velocity = ObjectsFem.makeConstraintInitialFlowVelocity(doc)
        self.assertTrue(is_derived_from(constraint_initial_flow_velocity, "App::DocumentObject"))
        self.assertTrue(is_derived_from(constraint_initial_flow_velocity, "Fem::ConstraintPython"))
        self.assertTrue(
            is_derived_from(constraint_initial_flow_velocity, "Fem::ConstraintInitialFlowVelocity")
        )

        # ConstraintInitialPressure
        constraint_initial_pressure = ObjectsFem.makeConstraintInitialPressure(doc)
        self.assertTrue(is_derived_from(constraint_initial_pressure, "App::DocumentObject"))
        self.assertTrue(is_derived_from(constraint_initial_pressure, "Fem::ConstraintPython"))
        self.assertTrue(
            is_derived_from(constraint_initial_pressure, "Fem::ConstraintInitialPressure")
        )

        # ConstraintInitialTemperature
        constraint_initial_temperature = ObjectsFem.makeConstraintInitialTemperature(doc)
        self.assertTrue(is_derived_from(constraint_initial_temperature, "App::DocumentObject"))
        self.assertTrue(is_derived_from(constraint_initial_temperature, "Fem::Constraint"))
        self.assertTrue(
            is_derived_from(constraint_initial_temperature, "Fem::ConstraintInitialTemperature")
        )

        # ConstraintPlaneRotation
        constraint_plane_rotation = ObjectsFem.makeConstraintPlaneRotation(doc)
        self.assertTrue(is_derived_from(constraint_plane_rotation, "App::DocumentObject"))
        self.assertTrue(is_derived_from(constraint_plane_rotation, "Fem::Constraint"))
        self.assertTrue(is_derived_from(constraint_plane_rotation, "Fem::ConstraintPlaneRotation"))

        # ConstraintPressure
        constraint_pressure = ObjectsFem.makeConstraintPressure(doc)
        self.assertTrue(is_derived_from(constraint_pressure, "App::DocumentObject"))
        self.assertTrue(is_derived_from(constraint_pressure, "Fem::Constraint"))
        self.assertTrue(is_derived_from(constraint_pressure, "Fem::ConstraintPressure"))

        # ConstraintPulley
        constraint_pulley = ObjectsFem.makeConstraintPulley(doc)
        self.assertTrue(is_derived_from(constraint_pulley, "App::DocumentObject"))
        self.assertTrue(is_derived_from(constraint_pulley, "Fem::Constraint"))
        self.assertTrue(is_derived_from(constraint_pulley, "Fem::ConstraintPulley"))

        # ConstraintSectionPrint
        constraint_self_weight = ObjectsFem.makeConstraintSectionPrint(doc)
        self.assertTrue(is_derived_from(constraint_self_weight, "App::DocumentObject"))
        self.assertTrue(is_derived_from(constraint_self_weight, "Fem::ConstraintPython"))
        self.assertTrue(is_derived_from(constraint_self_weight, "Fem::ConstraintSectionPrint"))

        # ConstraintSelfWeight
        constraint_self_weight = ObjectsFem.makeConstraintSelfWeight(doc)
        self.assertTrue(is_derived_from(constraint_self_weight, "App::DocumentObject"))
        self.assertTrue(is_derived_from(constraint_self_weight, "Fem::ConstraintPython"))
        self.assertTrue(is_derived_from(constraint_self_weight, "Fem::ConstraintSelfWeight"))

        # ConstraintCentrif
        constraint_centrif = ObjectsFem.makeConstraintCentrif(doc)
        self.assertTrue(is_derived_from(constraint_centrif, "App::DocumentObject"))
        self.assertTrue(is_derived_from(constraint_centrif, "Fem::ConstraintPython"))
        self.assertTrue(is_derived_from(constraint_centrif, "Fem::ConstraintCentrif"))

        # ConstraintTemperature
        constraint_temperature = ObjectsFem.makeConstraintTemperature(doc)
        self.assertTrue(is_derived_from(constraint_temperature, "App::DocumentObject"))
        self.assertTrue(is_derived_from(constraint_temperature, "Fem::Constraint"))
        self.assertTrue(is_derived_from(constraint_temperature, "Fem::ConstraintTemperature"))

        # ConstraintTie
        constraint_tie = ObjectsFem.makeConstraintTie(doc)
        self.assertTrue(is_derived_from(constraint_tie, "App::DocumentObject"))
        self.assertTrue(is_derived_from(constraint_tie, "Fem::Constraint"))
        self.assertTrue(is_derived_from(constraint_tie, "Fem::ConstraintPython"))
        self.assertTrue(is_derived_from(constraint_tie, "Fem::ConstraintTie"))

        # ConstraintTransform
        constraint_transform = ObjectsFem.makeConstraintTransform(doc)
        self.assertTrue(is_derived_from(constraint_transform, "App::DocumentObject"))
        self.assertTrue(is_derived_from(constraint_transform, "Fem::ConstraintTransform"))

        # ElementFluid1D
        fluid1d = ObjectsFem.makeElementFluid1D(doc)
        self.assertTrue(is_derived_from(fluid1d, "App::DocumentObject"))
        self.assertTrue(is_derived_from(fluid1d, "Fem::FeaturePython"))
        self.assertTrue(is_derived_from(fluid1d, "Fem::ElementFluid1D"))

        # ElementGeometry1D
        geometry1d = ObjectsFem.makeElementGeometry1D(doc)
        self.assertTrue(is_derived_from(geometry1d, "App::DocumentObject"))
        self.assertTrue(is_derived_from(geometry1d, "Fem::FeaturePython"))
        self.assertTrue(is_derived_from(geometry1d, "Fem::ElementGeometry1D"))

        # ElementGeometry2D
        geometry2d = ObjectsFem.makeElementGeometry2D(doc)
        self.assertTrue(is_derived_from(geometry2d, "App::DocumentObject"))
        self.assertTrue(is_derived_from(geometry2d, "Fem::FeaturePython"))
        self.assertTrue(is_derived_from(geometry2d, "Fem::ElementGeometry2D"))

        # ElementRotation1D
        rotation1d = ObjectsFem.makeElementRotation1D(doc)
        self.assertTrue(is_derived_from(rotation1d, "App::DocumentObject"))
        self.assertTrue(is_derived_from(rotation1d, "Fem::FeaturePython"))
        self.assertTrue(is_derived_from(rotation1d, "Fem::ElementRotation1D"))

        # Material Fluid
        material_fluid = ObjectsFem.makeMaterialFluid(doc)
        self.assertTrue(is_derived_from(material_fluid, "App::DocumentObject"))
        self.assertTrue(is_derived_from(material_fluid, "App::MaterialObjectPython"))
        self.assertTrue(is_derived_from(material_fluid, "Fem::MaterialCommon"))

        # Material Solid
        material_solid = ObjectsFem.makeMaterialSolid(doc)
        self.assertTrue(is_derived_from(material_solid, "App::DocumentObject"))
        self.assertTrue(is_derived_from(material_solid, "App::MaterialObjectPython"))
        self.assertTrue(is_derived_from(material_solid, "Fem::MaterialCommon"))

        # MaterialMechanicalNonlinear
        material_nonlinear = ObjectsFem.makeMaterialMechanicalNonlinear(doc, material_solid)
        self.assertTrue(is_derived_from(material_nonlinear, "App::DocumentObject"))
        self.assertTrue(is_derived_from(material_nonlinear, "Fem::FeaturePython"))
        self.assertTrue(is_derived_from(material_nonlinear, "Fem::MaterialMechanicalNonlinear"))

        # MaterialReinforced
        material_reinforced = ObjectsFem.makeMaterialReinforced(doc)
        self.assertTrue(is_derived_from(material_reinforced, "App::DocumentObject"))
        self.assertTrue(is_derived_from(material_reinforced, "App::MaterialObjectPython"))
        self.assertTrue(is_derived_from(material_reinforced, "Fem::MaterialReinforced"))

        # FemMeshGmsh
        mesh_gmsh = ObjectsFem.makeMeshGmsh(doc)
        self.assertTrue(is_derived_from(mesh_gmsh, "App::DocumentObject"))
        self.assertTrue(is_derived_from(mesh_gmsh, "Fem::FemMeshShapeBaseObjectPython"))
        self.assertTrue(is_derived_from(mesh_gmsh, "Fem::FemMeshGmsh"))

        # MeshBoundaryLayer
        mesh_boundarylayer = ObjectsFem.makeMeshBoundaryLayer(doc, mesh_gmsh)
        self.assertTrue(is_derived_from(mesh_boundarylayer, "App::DocumentObject"))
        self.assertTrue(is_derived_from(mesh_boundarylayer, "Fem::FeaturePython"))
        self.assertTrue(is_derived_from(mesh_boundarylayer, "Fem::MeshBoundaryLayer"))

        # MeshGroup
        mesh_group = ObjectsFem.makeMeshGroup(doc, mesh_gmsh)
        self.assertTrue(is_derived_from(mesh_group, "App::DocumentObject"))
        self.assertTrue(is_derived_from(mesh_group, "Fem::FeaturePython"))
        self.assertTrue(is_derived_from(mesh_group, "Fem::MeshGroup"))

        # MeshRegion
        mesh_region = ObjectsFem.makeMeshRegion(doc, mesh_gmsh)
        self.assertTrue(is_derived_from(mesh_region, "App::DocumentObject"))
        self.assertTrue(is_derived_from(mesh_region, "Fem::FeaturePython"))
        self.assertTrue(is_derived_from(mesh_region, "Fem::MeshRegion"))

        # FemMeshShapeNetgenObject
        mesh_netgen = ObjectsFem.makeMeshNetgen(doc)
        self.assertTrue(is_derived_from(mesh_netgen, "App::DocumentObject"))
        self.assertTrue(is_derived_from(mesh_netgen, "Fem::FemMeshShapeBaseObjectPython"))
        mesh_netgen = ObjectsFem.makeMeshNetgenLegacy(doc)
        self.assertTrue(is_derived_from(mesh_netgen, "App::DocumentObject"))
        self.assertTrue(is_derived_from(mesh_netgen, "Fem::FemMeshShapeNetgenObject"))

        # MeshResult
        mesh_result = ObjectsFem.makeMeshResult(doc)
        self.assertTrue(is_derived_from(mesh_result, "App::DocumentObject"))
        self.assertTrue(is_derived_from(mesh_result, "Fem::FemMeshObjectPython"))
        self.assertTrue(is_derived_from(mesh_result, "Fem::MeshResult"))

        # ResultMechanical
        result_mechanical = ObjectsFem.makeResultMechanical(doc)
        self.assertTrue(is_derived_from(result_mechanical, "App::DocumentObject"))
        self.assertTrue(is_derived_from(result_mechanical, "Fem::FemResultObjectPython"))
        self.assertTrue(is_derived_from(result_mechanical, "Fem::ResultMechanical"))

        # SolverCcxTools
        solver_ccxtools = ObjectsFem.makeSolverCalculiXCcxTools(doc)
        self.assertTrue(is_derived_from(solver_ccxtools, "App::DocumentObject"))
        self.assertTrue(is_derived_from(solver_ccxtools, "Fem::FemSolverObject"))
        self.assertTrue(is_derived_from(solver_ccxtools, "Fem::FemSolverObjectPython"))
        self.assertTrue(is_derived_from(solver_ccxtools, "Fem::SolverCcxTools"))

        # SolverCalculiX
        solver_calculix = ObjectsFem.makeSolverCalculiX(doc)
        self.assertTrue(is_derived_from(solver_calculix, "App::DocumentObject"))
        self.assertTrue(is_derived_from(solver_calculix, "Fem::FemSolverObject"))
        self.assertTrue(is_derived_from(solver_calculix, "Fem::FemSolverObjectPython"))
        self.assertTrue(is_derived_from(solver_calculix, "Fem::SolverCalculiX"))

        # SolverElmer
        solver_elmer = ObjectsFem.makeSolverElmer(doc)
        self.assertTrue(is_derived_from(solver_elmer, "App::DocumentObject"))
        self.assertTrue(is_derived_from(solver_elmer, "Fem::FemSolverObject"))
        self.assertTrue(is_derived_from(solver_elmer, "Fem::FemSolverObjectPython"))
        self.assertTrue(is_derived_from(solver_elmer, "Fem::SolverElmer"))

        # SolverMystran
        solver_mystran = ObjectsFem.makeSolverMystran(doc)
        self.assertTrue(is_derived_from(solver_mystran, "App::DocumentObject"))
        self.assertTrue(is_derived_from(solver_mystran, "Fem::FemSolverObject"))
        self.assertTrue(is_derived_from(solver_mystran, "Fem::FemSolverObjectPython"))
        self.assertTrue(is_derived_from(solver_mystran, "Fem::SolverMystran"))

        # SolverZ88
        solver_z88 = ObjectsFem.makeSolverZ88(doc)
        self.assertTrue(is_derived_from(solver_z88, "App::DocumentObject"))
        self.assertTrue(is_derived_from(solver_z88, "Fem::FemSolverObject"))
        self.assertTrue(is_derived_from(solver_z88, "Fem::FemSolverObjectPython"))
        self.assertTrue(is_derived_from(solver_z88, "Fem::SolverZ88"))

        # EquationElmerDeformation
        equation_deformation = ObjectsFem.makeEquationDeformation(doc, solver_elmer)
        self.assertTrue(is_derived_from(equation_deformation, "App::DocumentObject"))
        self.assertTrue(is_derived_from(equation_deformation, "App::FeaturePython"))
        self.assertTrue(is_derived_from(equation_deformation, "Fem::EquationElmerDeformation"))

        # EquationElmerElasticity
        equation_elasticity = ObjectsFem.makeEquationElasticity(doc, solver_elmer)
        self.assertTrue(is_derived_from(equation_elasticity, "App::DocumentObject"))
        self.assertTrue(is_derived_from(equation_elasticity, "App::FeaturePython"))
        self.assertTrue(is_derived_from(equation_elasticity, "Fem::EquationElmerElasticity"))

        # EquationElmerElectricforce
        equation_elasticity = ObjectsFem.makeEquationElectricforce(doc, solver_elmer)
        self.assertTrue(is_derived_from(equation_elasticity, "App::DocumentObject"))
        self.assertTrue(is_derived_from(equation_elasticity, "App::FeaturePython"))
        self.assertTrue(is_derived_from(equation_elasticity, "Fem::EquationElmerElectricforce"))

        # EquationElmerElectrostatic
        equation_electrostatic = ObjectsFem.makeEquationElectrostatic(doc, solver_elmer)
        self.assertTrue(is_derived_from(equation_electrostatic, "App::DocumentObject"))
        self.assertTrue(is_derived_from(equation_electrostatic, "App::FeaturePython"))
        self.assertTrue(is_derived_from(equation_electrostatic, "Fem::EquationElmerElectrostatic"))

        # EquationElmerFlow
        equation_flow = ObjectsFem.makeEquationFlow(doc, solver_elmer)
        self.assertTrue(is_derived_from(equation_flow, "App::DocumentObject"))
        self.assertTrue(is_derived_from(equation_flow, "App::FeaturePython"))
        self.assertTrue(is_derived_from(equation_flow, "Fem::EquationElmerFlow"))

        # EquationElmerFlux
        equation_flux = ObjectsFem.makeEquationFlux(doc, solver_elmer)
        self.assertTrue(is_derived_from(equation_flux, "App::DocumentObject"))
        self.assertTrue(is_derived_from(equation_flux, "App::FeaturePython"))
        self.assertTrue(is_derived_from(equation_flux, "Fem::EquationElmerFlux"))

        # EquationElmerHeat
        equation_heat = ObjectsFem.makeEquationHeat(doc, solver_elmer)
        self.assertTrue(is_derived_from(equation_heat, "App::DocumentObject"))
        self.assertTrue(is_derived_from(equation_heat, "App::FeaturePython"))
        self.assertTrue(is_derived_from(equation_heat, "Fem::EquationElmerHeat"))

        # EquationElmerMagnetodynamic2D
        equation_magnetodynamic2D = ObjectsFem.makeEquationMagnetodynamic2D(doc, solver_elmer)
        self.assertTrue(is_derived_from(equation_magnetodynamic2D, "App::DocumentObject"))
        self.assertTrue(is_derived_from(equation_magnetodynamic2D, "App::FeaturePython"))
        self.assertTrue(
            is_derived_from(equation_magnetodynamic2D, "Fem::EquationElmerMagnetodynamic2D")
        )

        # EquationElmerMagnetodynamic
        equation_magnetodynamic = ObjectsFem.makeEquationMagnetodynamic(doc, solver_elmer)
        self.assertTrue(is_derived_from(equation_magnetodynamic, "App::DocumentObject"))
        self.assertTrue(is_derived_from(equation_magnetodynamic, "App::FeaturePython"))
        self.assertTrue(
            is_derived_from(equation_magnetodynamic, "Fem::EquationElmerMagnetodynamic")
        )

        # EquationElmerStaticCurrent
        equation_staticcurrent = ObjectsFem.makeEquationStaticCurrent(doc, solver_elmer)
        self.assertTrue(is_derived_from(equation_staticcurrent, "App::DocumentObject"))
        self.assertTrue(is_derived_from(equation_staticcurrent, "App::FeaturePython"))
        self.assertTrue(is_derived_from(equation_staticcurrent, "Fem::EquationElmerStaticCurrent"))

        fcc_print(
            "doc objects count: {}, method: {}".format(
                len(doc.Objects), sys._getframe().f_code.co_name
            )
        )
        # TODO: vtk post objs, thus 5 obj less than test_femobjects_make
        self.assertEqual(len(doc.Objects), testtools.get_defmake_count(False))
        # TODO constraint transform is not derived from "Fem::Constraint" ?!?

    # ********************************************************************************************
    def test_femobjects_derivedfromstd(self):
        # only the last True type is used
        doc = self.document

        self.assertTrue(ObjectsFem.makeAnalysis(doc).isDerivedFrom("Fem::FemAnalysis"))
        self.assertTrue(
            ObjectsFem.makeConstantVacuumPermittivity(doc).isDerivedFrom("Fem::ConstraintPython")
        )
        self.assertTrue(
            ObjectsFem.makeConstraintBearing(doc).isDerivedFrom("Fem::ConstraintBearing")
        )
        self.assertTrue(
            ObjectsFem.makeConstraintBodyHeatSource(doc).isDerivedFrom("Fem::ConstraintPython")
        )
        self.assertTrue(
            ObjectsFem.makeConstraintContact(doc).isDerivedFrom("Fem::ConstraintContact")
        )
        self.assertTrue(
            ObjectsFem.makeConstraintCurrentDensity(doc).isDerivedFrom("Fem::ConstraintPython")
        )
        self.assertTrue(
            ObjectsFem.makeConstraintDisplacement(doc).isDerivedFrom("Fem::ConstraintDisplacement")
        )
        self.assertTrue(
            ObjectsFem.makeConstraintElectrostaticPotential(doc).isDerivedFrom(
                "Fem::ConstraintPython"
            )
        )
        self.assertTrue(
            ObjectsFem.makeConstraintElectricChargeDensity(doc).isDerivedFrom(
                "Fem::ConstraintPython"
            )
        )
        self.assertTrue(ObjectsFem.makeConstraintFixed(doc).isDerivedFrom("Fem::ConstraintFixed"))
        self.assertTrue(
            ObjectsFem.makeConstraintRigidBody(doc).isDerivedFrom("Fem::ConstraintRigidBody")
        )
        self.assertTrue(
            ObjectsFem.makeConstraintFlowVelocity(doc).isDerivedFrom("Fem::ConstraintPython")
        )
        self.assertTrue(
            ObjectsFem.makeConstraintFluidBoundary(doc).isDerivedFrom(
                "Fem::ConstraintFluidBoundary"
            )
        )
        self.assertTrue(
            ObjectsFem.makeConstraintMagnetization(doc).isDerivedFrom("Fem::ConstraintPython")
        )
        self.assertTrue(ObjectsFem.makeConstraintSpring(doc).isDerivedFrom("Fem::ConstraintSpring"))
        self.assertTrue(ObjectsFem.makeConstraintForce(doc).isDerivedFrom("Fem::ConstraintForce"))
        self.assertTrue(ObjectsFem.makeConstraintGear(doc).isDerivedFrom("Fem::ConstraintGear"))
        self.assertTrue(
            ObjectsFem.makeConstraintHeatflux(doc).isDerivedFrom("Fem::ConstraintHeatflux")
        )
        self.assertTrue(
            ObjectsFem.makeConstraintInitialFlowVelocity(doc).isDerivedFrom("Fem::ConstraintPython")
        )
        self.assertTrue(
            ObjectsFem.makeConstraintInitialPressure(doc).isDerivedFrom("Fem::ConstraintPython")
        )
        self.assertTrue(
            ObjectsFem.makeConstraintInitialTemperature(doc).isDerivedFrom(
                "Fem::ConstraintInitialTemperature"
            )
        )
        self.assertTrue(
            ObjectsFem.makeConstraintPlaneRotation(doc).isDerivedFrom(
                "Fem::ConstraintPlaneRotation"
            )
        )
        self.assertTrue(
            ObjectsFem.makeConstraintPressure(doc).isDerivedFrom("Fem::ConstraintPressure")
        )
        self.assertTrue(ObjectsFem.makeConstraintPulley(doc).isDerivedFrom("Fem::ConstraintPulley"))
        self.assertTrue(
            ObjectsFem.makeConstraintSectionPrint(doc).isDerivedFrom("Fem::ConstraintPython")
        )
        self.assertTrue(
            ObjectsFem.makeConstraintSelfWeight(doc).isDerivedFrom("Fem::ConstraintPython")
        )
        self.assertTrue(
            ObjectsFem.makeConstraintCentrif(doc).isDerivedFrom("Fem::ConstraintPython")
        )
        self.assertTrue(
            ObjectsFem.makeConstraintTemperature(doc).isDerivedFrom("Fem::ConstraintTemperature")
        )
        self.assertTrue(ObjectsFem.makeConstraintTie(doc).isDerivedFrom("Fem::ConstraintPython"))
        self.assertTrue(
            ObjectsFem.makeConstraintTransform(doc).isDerivedFrom("Fem::ConstraintTransform")
        )
        self.assertTrue(ObjectsFem.makeElementFluid1D(doc).isDerivedFrom("Fem::FeaturePython"))
        self.assertTrue(ObjectsFem.makeElementGeometry1D(doc).isDerivedFrom("Fem::FeaturePython"))
        self.assertTrue(ObjectsFem.makeElementGeometry2D(doc).isDerivedFrom("Fem::FeaturePython"))
        self.assertTrue(ObjectsFem.makeElementRotation1D(doc).isDerivedFrom("Fem::FeaturePython"))
        materialsolid = ObjectsFem.makeMaterialSolid(doc)
        self.assertTrue(
            ObjectsFem.makeMaterialFluid(doc).isDerivedFrom("App::MaterialObjectPython")
        )
        self.assertTrue(materialsolid.isDerivedFrom("App::MaterialObjectPython"))
        self.assertTrue(
            ObjectsFem.makeMaterialMechanicalNonlinear(doc, materialsolid).isDerivedFrom(
                "Fem::FeaturePython"
            )
        )
        self.assertTrue(
            ObjectsFem.makeMaterialReinforced(doc).isDerivedFrom("App::MaterialObjectPython")
        )
        mesh = ObjectsFem.makeMeshGmsh(doc)
        self.assertTrue(mesh.isDerivedFrom("Fem::FemMeshShapeBaseObjectPython"))
        self.assertTrue(
            ObjectsFem.makeMeshBoundaryLayer(doc, mesh).isDerivedFrom("Fem::FeaturePython")
        )
        self.assertTrue(ObjectsFem.makeMeshGroup(doc, mesh).isDerivedFrom("Fem::FeaturePython"))
        self.assertTrue(ObjectsFem.makeMeshRegion(doc, mesh).isDerivedFrom("Fem::FeaturePython"))
        self.assertTrue(
            ObjectsFem.makeMeshNetgen(doc).isDerivedFrom("Fem::FemMeshShapeBaseObjectPython")
        )
        self.assertTrue(
            ObjectsFem.makeMeshNetgenLegacy(doc).isDerivedFrom("Fem::FemMeshShapeNetgenObject")
        )
        self.assertTrue(ObjectsFem.makeMeshResult(doc).isDerivedFrom("Fem::FemMeshObjectPython"))
        self.assertTrue(
            ObjectsFem.makeResultMechanical(doc).isDerivedFrom("Fem::FemResultObjectPython")
        )
        solverelmer = ObjectsFem.makeSolverElmer(doc)
        self.assertTrue(
            ObjectsFem.makeSolverCalculiXCcxTools(doc).isDerivedFrom("Fem::FemSolverObjectPython")
        )
        self.assertTrue(
            ObjectsFem.makeSolverCalculiX(doc).isDerivedFrom("Fem::FemSolverObjectPython")
        )
        self.assertTrue(solverelmer.isDerivedFrom("Fem::FemSolverObjectPython"))
        self.assertTrue(
            ObjectsFem.makeSolverMystran(doc).isDerivedFrom("Fem::FemSolverObjectPython")
        )
        self.assertTrue(ObjectsFem.makeSolverZ88(doc).isDerivedFrom("Fem::FemSolverObjectPython"))
        self.assertTrue(
            ObjectsFem.makeEquationDeformation(doc, solverelmer).isDerivedFrom("App::FeaturePython")
        )
        self.assertTrue(
            ObjectsFem.makeEquationElasticity(doc, solverelmer).isDerivedFrom("App::FeaturePython")
        )
        self.assertTrue(
            ObjectsFem.makeEquationElectricforce(doc, solverelmer).isDerivedFrom(
                "App::FeaturePython"
            )
        )
        self.assertTrue(
            ObjectsFem.makeEquationElectrostatic(doc, solverelmer).isDerivedFrom(
                "App::FeaturePython"
            )
        )
        self.assertTrue(
            ObjectsFem.makeEquationFlow(doc, solverelmer).isDerivedFrom("App::FeaturePython")
        )
        self.assertTrue(
            ObjectsFem.makeEquationFlux(doc, solverelmer).isDerivedFrom("App::FeaturePython")
        )
        self.assertTrue(
            ObjectsFem.makeEquationHeat(doc, solverelmer).isDerivedFrom("App::FeaturePython")
        )
        self.assertTrue(
            ObjectsFem.makeEquationMagnetodynamic2D(doc, solverelmer).isDerivedFrom(
                "App::FeaturePython"
            )
        )
        self.assertTrue(
            ObjectsFem.makeEquationMagnetodynamic(doc, solverelmer).isDerivedFrom(
                "App::FeaturePython"
            )
        )
        self.assertTrue(
            ObjectsFem.makeEquationStaticCurrent(doc, solverelmer).isDerivedFrom(
                "App::FeaturePython"
            )
        )

        fcc_print(
            "doc objects count: {}, method: {}".format(
                len(doc.Objects), sys._getframe().f_code.co_name
            )
        )
        # TODO: vtk post objs, thus 5 obj less than test_femobjects_make
        self.assertEqual(len(doc.Objects), testtools.get_defmake_count(False))


# helper
def create_all_fem_objects_doc(doc):
    analysis = ObjectsFem.makeAnalysis(doc)

    analysis.addObject(ObjectsFem.makeConstantVacuumPermittivity(doc))
    analysis.addObject(ObjectsFem.makeConstraintBearing(doc))
    analysis.addObject(ObjectsFem.makeConstraintBodyHeatSource(doc))
    analysis.addObject(ObjectsFem.makeConstraintContact(doc))
    analysis.addObject(ObjectsFem.makeConstraintCurrentDensity(doc))
    analysis.addObject(ObjectsFem.makeConstraintDisplacement(doc))
    analysis.addObject(ObjectsFem.makeConstraintElectrostaticPotential(doc))
    analysis.addObject(ObjectsFem.makeConstraintElectricChargeDensity(doc))
    analysis.addObject(ObjectsFem.makeConstraintFixed(doc))
    analysis.addObject(ObjectsFem.makeConstraintRigidBody(doc))
    analysis.addObject(ObjectsFem.makeConstraintFlowVelocity(doc))
    analysis.addObject(ObjectsFem.makeConstraintFluidBoundary(doc))
    analysis.addObject(ObjectsFem.makeConstraintSpring(doc))
    analysis.addObject(ObjectsFem.makeConstraintForce(doc))
    analysis.addObject(ObjectsFem.makeConstraintGear(doc))
    analysis.addObject(ObjectsFem.makeConstraintHeatflux(doc))
    analysis.addObject(ObjectsFem.makeConstraintInitialFlowVelocity(doc))
    analysis.addObject(ObjectsFem.makeConstraintInitialPressure(doc))
    analysis.addObject(ObjectsFem.makeConstraintInitialTemperature(doc))
    analysis.addObject(ObjectsFem.makeConstraintMagnetization(doc))
    analysis.addObject(ObjectsFem.makeConstraintPlaneRotation(doc))
    analysis.addObject(ObjectsFem.makeConstraintPressure(doc))
    analysis.addObject(ObjectsFem.makeConstraintPulley(doc))
    analysis.addObject(ObjectsFem.makeConstraintSectionPrint(doc))
    analysis.addObject(ObjectsFem.makeConstraintSelfWeight(doc))
    analysis.addObject(ObjectsFem.makeConstraintCentrif(doc))
    analysis.addObject(ObjectsFem.makeConstraintTemperature(doc))
    analysis.addObject(ObjectsFem.makeConstraintTie(doc))
    analysis.addObject(ObjectsFem.makeConstraintTransform(doc))

    analysis.addObject(ObjectsFem.makeElementFluid1D(doc))
    analysis.addObject(ObjectsFem.makeElementGeometry1D(doc))
    analysis.addObject(ObjectsFem.makeElementGeometry2D(doc))
    analysis.addObject(ObjectsFem.makeElementRotation1D(doc))

    analysis.addObject(ObjectsFem.makeMaterialFluid(doc))
    mat = analysis.addObject(ObjectsFem.makeMaterialSolid(doc))[0]
    analysis.addObject(ObjectsFem.makeMaterialMechanicalNonlinear(doc, mat))
    analysis.addObject(ObjectsFem.makeMaterialReinforced(doc))

    msh = analysis.addObject(ObjectsFem.makeMeshGmsh(doc))[0]
    ObjectsFem.makeMeshBoundaryLayer(doc, msh)
    ObjectsFem.makeMeshGroup(doc, msh)
    ObjectsFem.makeMeshRegion(doc, msh)
    analysis.addObject(ObjectsFem.makeMeshNetgen(doc))
    analysis.addObject(ObjectsFem.makeMeshNetgenLegacy(doc))
    rm = ObjectsFem.makeMeshResult(doc)

    res = analysis.addObject(ObjectsFem.makeResultMechanical(doc))[0]
    res.Mesh = rm
    if "BUILD_FEM_VTK" in FreeCAD.__cmake__:
        vres = analysis.addObject(ObjectsFem.makePostVtkResult(doc, [res]))[0]
        ObjectsFem.makePostVtkFilterClipRegion(doc, vres)
        ObjectsFem.makePostVtkFilterClipScalar(doc, vres)
        ObjectsFem.makePostVtkFilterCutFunction(doc, vres)
        ObjectsFem.makePostVtkFilterWarp(doc, vres)
        ObjectsFem.makePostVtkFilterContours(doc, vres)
        if "BUILD_FEM_VTK_PYTHON" in FreeCAD.__cmake__:
            ObjectsFem.makePostFilterGlyph(doc, vres)

            # data extraction objects
            lp = analysis.addObject(ObjectsFem.makePostLineplot(doc))[0]
            lp.addObject(ObjectsFem.makePostLineplotFieldData(doc))
            lp.addObject(ObjectsFem.makePostLineplotIndexOverFrames(doc))

            hp = analysis.addObject(ObjectsFem.makePostHistogram(doc))[0]
            hp.addObject(ObjectsFem.makePostHistogramFieldData(doc))
            hp.addObject(ObjectsFem.makePostHistogramIndexOverFrames(doc))

            tb = analysis.addObject(ObjectsFem.makePostTable(doc))[0]
            tb.addObject(ObjectsFem.makePostTableFieldData(doc))
            tb.addObject(ObjectsFem.makePostTableIndexOverFrames(doc))

    analysis.addObject(ObjectsFem.makeSolverCalculiXCcxTools(doc))
    analysis.addObject(ObjectsFem.makeSolverCalculiX(doc))
    sol = analysis.addObject(ObjectsFem.makeSolverElmer(doc))[0]
    analysis.addObject(ObjectsFem.makeSolverMystran(doc))
    analysis.addObject(ObjectsFem.makeSolverZ88(doc))

    ObjectsFem.makeEquationDeformation(doc, sol)
    ObjectsFem.makeEquationElasticity(doc, sol)
    ObjectsFem.makeEquationElectricforce(doc, sol)
    ObjectsFem.makeEquationElectrostatic(doc, sol)
    ObjectsFem.makeEquationFlow(doc, sol)
    ObjectsFem.makeEquationFlux(doc, sol)
    ObjectsFem.makeEquationHeat(doc, sol)
    ObjectsFem.makeEquationMagnetodynamic2D(doc, sol)
    ObjectsFem.makeEquationMagnetodynamic(doc, sol)
    ObjectsFem.makeEquationStaticCurrent(doc, sol)

    doc.recompute()

    return doc



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>                  *
# *   (c) 2009, 2010 Ken Cline <cline@frii.com>                             *
# *   (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides GUI tools to create Circle objects."""
## @package gui_circles
# \ingroup draftguitools
# \brief Provides GUI tools to create Circle objects.

## \addtogroup draftguitools
# @{
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCADGui as Gui
import Draft_rc
import draftguitools.gui_arcs as gui_arcs

# The module is used to prevent complaints from code checkers (flake8)
True if Draft_rc.__name__ else False


class Circle(gui_arcs.Arc):
    """Gui command for the Circle tool.

    It inherits the entire `Arc` class.
    The only difference is that the `closedCircle` attribute
    is already set to `True`, and the `featureName` attribute
    is `'Circle'`.

    This will result in an arc that describes a complete circumference
    so the starting angle and end angle will be the same.

    Internally, both circular arcs and circles are `'Circle'` objects.

    Discussion
    ----------
    Both arcs and circles are `'Circle'` objects, but when it comes to the
    Gui Commands, the relationships are reversed, and both launch the `Arc`
    command.

    Maybe the relationship should be changed: the base Gui Command
    should be `Circle`, and an arc would launch the same command,
    as both are internally `'Circle'` objects.

    Another possibility is to rename the `'Circle'` object to `'Arc'`.
    Then both a circle and an arc would internally be `'Arc'` objects,
    and in the Gui Commands they both would use the `Arc` command.
    """

    def __init__(self):
        super().__init__()
        self.closedCircle = True
        self.featureName = "Circle"

    def GetResources(self):
        """Set icon, menu and tooltip."""

        return {
            "Pixmap": "Draft_Circle",
            "Accel": "C, I",
            "MenuText": QT_TRANSLATE_NOOP("Draft_Circle", "Circle"),
            "ToolTip": QT_TRANSLATE_NOOP("Draft_Circle", "Creates a circle (full circular arc)"),
        }


Gui.addCommand("Draft_Circle", Circle())

## @}



# ***************************************************************************
# *   Copyright (c) 2011 Juergen Riegel <FreeCAD@juergen-riegel.net>        *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

from __future__ import division
from math import pi
import unittest

import FreeCAD


class TestFillet(unittest.TestCase):
    def setUp(self):
        self.Doc = FreeCAD.newDocument("PartDesignTestFillet")

    def testFilletCubeToSphere(self):
        self.Body = self.Doc.addObject("PartDesign::Body", "Body")
        self.Box = self.Doc.addObject("PartDesign::AdditiveBox", "Box")
        self.Body.addObject(self.Box)
        self.Box.Length = 10.00
        self.Box.Width = 10.00
        self.Box.Height = 10.00
        self.Doc.recompute()
        self.Fillet = self.Doc.addObject("PartDesign::Fillet", "Fillet")
        self.Fillet.Base = (self.Box, ["Face" + str(i + 1) for i in range(6)])
        self.Fillet.Radius = 4.999999
        self.Body.addObject(self.Fillet)
        self.Doc.recompute()
        self.assertAlmostEqual(self.Fillet.Shape.Volume, 4 / 3 * pi * 5**3, places=3)
        # test UseAllEdges property
        self.Fillet.UseAllEdges = True
        self.Fillet.Base = (self.Box, [""])  # no subobjects, should still work
        self.Doc.recompute()
        self.assertAlmostEqual(self.Fillet.Shape.Volume, 4 / 3 * pi * 5**3, places=3)
        self.Fillet.Base = (self.Box, ["Face50"])  # non-existent face, topo naming resilience
        self.Doc.recompute()
        self.assertAlmostEqual(self.Fillet.Shape.Volume, 4 / 3 * pi * 5**3, places=3)
        self.Fillet.UseAllEdges = False
        self.Fillet.Base = (self.Box, ["Face1"])
        self.Doc.recompute()
        self.assertNotAlmostEqual(self.Fillet.Shape.Volume, 4 / 3 * pi * 5**3, places=3)

    def tearDown(self):
        # closing doc
        FreeCAD.closeDocument("PartDesignTestFillet")
        # print ("omit closing document for debugging")



# SPDX-License-Identifier: LGPL-2.1-or-later

import unittest

import FreeCAD as App


class ColorTransparencyTest(unittest.TestCase):

    def setUp(self):
        self._doc = App.newDocument()
        self._pg = App.ParamGet("User parameter:BaseApp/Preferences/View")
        self._backup_default_transparency = self._pg.GetInt("DefaultShapeTransparency")
        self._backup_default_shapecolor = self._pg.GetUnsigned("DefaultShapeColor")

    def tearDown(self):
        App.closeDocument(self._doc.Name)
        self._pg.SetInt("DefaultShapeTransparency", self._backup_default_transparency)
        self._pg.SetUnsigned("DefaultShapeColor", self._backup_default_shapecolor)

    def test_default_shape_transparency(self):
        """
        related: https://github.com/FreeCAD/FreeCAD/pull/11866
        related: https://github.com/FreeCAD/FreeCAD/pull/11586
        """
        transparency = 70
        self._pg.SetInt("DefaultShapeTransparency", transparency)
        obj = self._doc.addObject("Part::Box")
        assert obj.ViewObject.Transparency == transparency
        obj.ViewObject.ShapeAppearance[0].DiffuseColor = (0.5, 0.0, 0.0)

        self.assertEqual(
            obj.ViewObject.Transparency,
            transparency,
            "transparency was unexpectedly changed to {} when changing the color.".format(
                obj.ViewObject.Transparency
            ),
        )

    def test_default_shape_color(self):
        """
        related: https://github.com/FreeCAD/FreeCAD/pull/11866
        """

        """
        This test isn't currently valid as it draws from the hard coded default material.

        The preference editor doesn't allow for setting transparencies. The default value
        of 0 corresponds to a fully transparent color, which is not desirable. It changes
        the transparency when loading to 1.0
        """

        self._pg.SetUnsigned("DefaultShapeColor", 0xFF000000)  # red
        obj = self._doc.addObject("Part::Box")

        self.assertEqual(
            obj.ViewObject.ShapeAppearance[0].DiffuseColor,
            (1.0, 0.0, 0.0, 1.0),
            "default shape color was not set correctly",
        )
        self.assertEqual(
            obj.ViewObject.ShapeMaterial.DiffuseColor,
            (1.0, 0.0, 0.0, 1.0),
            "default material color was not set correctly",
        )

    def test_app_plane_transparency(self):
        """
        related: https://github.com/FreeCAD/FreeCAD/pull/12064
        """
        self._pg.SetInt("DefaultShapeTransparency", 70)
        obj = self._doc.addObject("App::Origin")
        t = self._doc.findObjects("App::Plane")[0].ViewObject.Transparency

        self.assertEqual(t, 0, "transparency of App::Plane object is {} instead of 0".format(t))



# ***************************************************************************
# *   Copyright (c) 2023 edi <edi271@a1.net>                                *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides several TechDraw GuiCommands to create vertexes."""

__title__ = "TechDrawTools.CommandVertexCreations"
__author__ = "edi"
__url__ = "https://www.freecad.org"
__version__ = "00.01"
__date__ = "2023/12/05"


from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import FreeCADGui as Gui

import TechDrawTools
import TechDrawTools.TDToolsUtil as Utils


import TechDraw

class CommandVertexCreationGroup:
    '''Create a drop down toolbar/menubar for vertex creating tools'''
    def Activated(self, index):
        if index == 0:
            Gui.runCommand("TechDraw_ExtensionVertexAtIntersection")
        elif index == 1:
            Gui.runCommand("TechDraw_CommandAddOffsetVertex")

    def GetCommands(self):
        return("TechDraw_ExtensionVertexAtIntersection",
               "TechDraw_CommandAddOffsetVertex")

    def GetDefaultCommand(self):
        return 0

    def GetResources(self):
        """Return a dictionary with data that will be used by the button or menu item."""
        return {'Pixmap': 'TechDraw_ExtensionVertexAtIntersection.svg',
                'Accel': "",
                'MenuText': QT_TRANSLATE_NOOP("TechDraw_ExtensionVertexAtIntersection","Cosmetic Intersection Vertices"),
                'ToolTip': QT_TRANSLATE_NOOP("TechDraw_ExtensionVertexAtIntersection", "Adds cosmetic vertices at the intersectionss of selected edges")}

    def IsActive(self):
        """Return True when the command should be active or False when it should be disabled (greyed)."""
        if App.ActiveDocument:
            return Utils.havePage() and Utils.haveView()
        else:
            return False

class CommandAddOffsetVertex:
    """Creates a vertex offset to a selected vertex."""

    def __init__(self):
        """Initialize variables for the command that must exist at all times."""
        pass

    def GetResources(self):
        """Return a dictionary with data that will be used by the button or menu item."""
        return {'Pixmap': 'actions/TechDraw_AddOffsetVertex.svg',
                'Accel': "",
                'MenuText': QT_TRANSLATE_NOOP("TechDraw_AddOffsetVertex", "Offset Vertex"),
                'ToolTip': QT_TRANSLATE_NOOP("TechDraw_AddOffsetVertex", "Creates an offset from one selected vertex")}

    def Activated(self):
        """Run the following code when the command is activated (button pressed)."""
        if Utils.getSelView() and Utils.getSelVertexes():
            view = Utils.getSelView()
            vertexes = Utils.getSelVertexes()
            self.ui = TechDrawTools.TaskAddOffsetVertex(view, vertexes[0])
            Gui.Control.showDialog(self.ui)

    def IsActive(self):
        """Return True when the command should be active or False when it should be disabled (greyed)."""
        if App.ActiveDocument:
            return Utils.havePage() and Utils.haveView()
        else:
            return False

Gui.addCommand('TechDraw_CommandVertexCreationGroup',CommandVertexCreationGroup())
Gui.addCommand('TechDraw_CommandAddOffsetVertex',CommandAddOffsetVertex())
